@@ -18,7 +18,7 @@
  */
 
 /***************************************************************************
- * Copyright (C) 2013-2015 Ping Identity Corporation
+ * Copyright (C) 2013-2016 Ping Identity Corporation
  * All rights reserved.
  *
  * For further information please contact:
@@ -52,6 +52,7 @@
 #include <apr_errno.h>
 #include <apr_strings.h>
 #include <apr_portable.h>
+#include <apr_base64.h>
 
 #include <httpd.h>
 #include <http_config.h>
@@ -130,6 +131,8 @@
 #define OIDC_DEFAULT_OAUTH_EXPIRY_CLAIM_FORMAT "relative"
 /* default OAuth 2.0 non-spec compliant introspection expiry claim required */
 #define OIDC_DEFAULT_OAUTH_EXPIRY_CLAIM_REQUIRED TRUE
+/* default refresh interval in seconds after which claims from the user info endpoint should be refreshed */
+#define OIDC_DEFAULT_USERINFO_REFRESH_INTERVAL 0
 
 extern module AP_MODULE_DECLARE_DATA auth_openidc_module;
 
@@ -288,9 +291,22 @@ static const char *oidc_set_session_type(cmd_parms *cmd, void *ptr,
 	oidc_cfg *cfg = (oidc_cfg *) ap_get_module_config(
 			cmd->server->module_config, &auth_openidc_module);
 
-	if (apr_strnatcmp(arg, "server-cache") == 0) {
+	char *s = apr_pstrdup(cmd->pool, arg);
+	char *p = strstr(s, ":");
+
+	if (p) {
+		if (apr_strnatcmp(p, ":persistent") != 0) {
+			return (apr_psprintf(cmd->pool,
+					"oidc_set_session_type: invalid suffix value for %s (%s); only \":persistent\" is allowed",
+					cmd->directive->directive, arg));
+		}
+		cfg->persistent_session_cookie = 1;
+		*p = '\0';
+	}
+
+	if (apr_strnatcmp(s, "server-cache") == 0) {
 		cfg->session_type = OIDC_SESSION_TYPE_22_SERVER_CACHE;
-	} else if (apr_strnatcmp(arg, "client-cookie") == 0) {
+	} else if (apr_strnatcmp(s, "client-cookie") == 0) {
 		cfg->session_type = OIDC_SESSION_TYPE_22_CLIENT_COOKIE;
 	} else {
 		return (apr_psprintf(cmd->pool,
@@ -364,11 +380,13 @@ static const char *oidc_set_endpoint_auth_slot(cmd_parms *cmd, void *struct_ptr,
 			cmd->server->module_config, &auth_openidc_module);
 
 	if ((apr_strnatcmp(arg, "client_secret_post") == 0)
-			|| (apr_strnatcmp(arg, "client_secret_basic") == 0)) {
+			|| (apr_strnatcmp(arg, "client_secret_basic") == 0)
+			|| (apr_strnatcmp(arg, "client_secret_jwt") == 0)
+			|| (apr_strnatcmp(arg, "private_key_jwt") == 0)) {
 
 		return ap_set_string_slot(cmd, cfg, arg);
 	}
-	return "parameter must be 'client_secret_post' or 'client_secret_basic'";
+	return "parameter must be 'client_secret_post', 'client_secret_basic', 'client_secret_jwt' or 'private_key_jwt'";
 }
 
 /*
@@ -388,6 +406,20 @@ static const char *oidc_set_response_type(cmd_parms *cmd, void *struct_ptr,
 					'|'));
 }
 
+/*
+ * define the PCKE method to use
+ */
+static const char *oidc_set_pkce_method(cmd_parms *cmd, void *ptr,
+		const char *arg) {
+	oidc_cfg *cfg = (oidc_cfg *) ap_get_module_config(
+			cmd->server->module_config, &auth_openidc_module);
+
+	if ((apr_strnatcmp(arg, "plain") == 0) || (apr_strnatcmp(arg, "S256") == 0))
+		return ap_set_string_slot(cmd, cfg, arg);
+
+	return "parameter must be 'plain' or 'S256'";
+}
+
 /*
  * set the response mode used
  */
@@ -516,6 +548,86 @@ static const char *oidc_set_session_max_duration(cmd_parms *cmd,
 	return NULL;
 }
 
+/*
+ * decode a key value based on the provided encoding b64|b64url|hex|plain
+ */
+static char * oidc_config_decode_key(apr_pool_t *pool, const char *enc,
+		const char *input, char **key, int *key_len) {
+	if (apr_strnatcmp(enc, "b64") == 0) {
+		int len = apr_base64_decode_len(input);
+		*key = apr_palloc(pool, len);
+		*key_len = apr_base64_decode(*key, input);
+		if (key_len <= 0) {
+			return "base64 decoding failed";
+		}
+		return NULL;
+	} else if (apr_strnatcmp(enc, "b64url") == 0) {
+		*key_len = apr_jwt_base64url_decode(pool, key, input, 1);
+		return NULL;
+	} else if (apr_strnatcmp(enc, "hex") == 0) {
+		*key_len = strlen(input) / 2;
+		const char *pos = input;
+		unsigned char *val = apr_palloc(pool, *key_len);
+		size_t count = 0;
+		for (count = 0; count < (*key_len) / sizeof(unsigned char); count++) {
+			sscanf(pos, "%2hhx", &val[count]);
+			pos += 2;
+		}
+		*key = (char*) val;
+		return NULL;
+	} else if (apr_strnatcmp(enc, "plain") == 0) {
+		*key = apr_pstrdup(pool, input);
+		*key_len = strlen(*key);
+		return NULL;
+	}
+	return apr_psprintf(pool,
+			"unknown encoding: %s; must be one of b64|b64url|hex|plain", enc);
+}
+
+#define OIDC_KEY_TUPLE_SEPARATOR "#"
+
+/*
+ * set a <encoding>#<key-identifier>#<key> tuple
+ */
+static char * oidc_config_get_id_key_tuple(apr_pool_t *pool, const char *tuple,
+		char **kid, char **key, int *key_len, apr_byte_t triplet) {
+	char *s = NULL, *p = NULL, *q = NULL;
+	char *enc = NULL, *err = NULL;
+
+	if ((tuple == NULL) || (apr_strnatcmp(tuple, "") == 0))
+		return "tuple value not set";
+
+	s = apr_pstrdup(pool, tuple);
+	p = strstr(s, OIDC_KEY_TUPLE_SEPARATOR);
+	if (p && triplet)
+		q = strstr(p + 1, OIDC_KEY_TUPLE_SEPARATOR);
+
+	if (p) {
+		if (q) {
+			*p = '\0';
+			*q = '\0';
+			enc = s;
+			p++;
+			if (p != q)
+				*kid = apr_pstrdup(pool, p);
+			err = oidc_config_decode_key(pool, enc, q + 1, key, key_len);
+			if (err)
+				return err;
+		} else {
+			*p = '\0';
+			*kid = s;
+			*key = p + 1;
+			*key_len = strlen(*key);
+		}
+	} else {
+		*kid = NULL;
+		*key = s;
+		*key_len = strlen(*key);
+	}
+
+	return NULL;
+}
+
 /*
  * add a public key from an X.509 file to our list of JWKs with public keys
  */
@@ -530,10 +642,17 @@ static const char *oidc_set_public_key_files(cmd_parms *cmd, void *struct_ptr,
 	int offset = (int) (long) cmd->info;
 	apr_hash_t **public_keys = (apr_hash_t **) ((char *) cfg + offset);
 
-	if (apr_jwk_parse_rsa_public_key(cmd->pool, arg, &jwk, &err) == FALSE) {
+	char *kid = NULL, *fname = NULL;
+	int fname_len;
+	char *rv = oidc_config_get_id_key_tuple(cmd->pool, arg, &kid, &fname,
+			&fname_len, FALSE);
+	if (rv != NULL)
+		return rv;
+
+	if (apr_jwk_parse_rsa_public_key(cmd->pool, kid, fname, &jwk, &err) == FALSE) {
 		return apr_psprintf(cmd->pool,
-				"apr_jwk_parse_rsa_public_key failed for \"%s\": %s", arg,
-				apr_jwt_e2s(cmd->pool, err));
+				"apr_jwk_parse_rsa_public_key failed for (kid=%s) \"%s\": %s",
+				kid, fname, apr_jwt_e2s(cmd->pool, err));
 	}
 
 	if (*public_keys == NULL)
@@ -548,12 +667,33 @@ static const char *oidc_set_public_key_files(cmd_parms *cmd, void *struct_ptr,
  */
 static const char *oidc_set_shared_keys(cmd_parms *cmd, void *struct_ptr,
 		const char *arg) {
+	apr_jwt_error_t err;
+	apr_jwk_t *jwk = NULL;
+
 	oidc_cfg *cfg = (oidc_cfg *) ap_get_module_config(
 			cmd->server->module_config, &auth_openidc_module);
 	int offset = (int) (long) cmd->info;
 	apr_hash_t **shared_keys = (apr_hash_t **) ((char *) cfg + offset);
-	*shared_keys = oidc_util_merge_symmetric_key(cmd->pool, *shared_keys, arg,
-			NULL);
+
+	char *kid = NULL, *secret = NULL;
+	int key_len = 0;
+	char *rv = oidc_config_get_id_key_tuple(cmd->pool, arg, &kid, &secret,
+			&key_len, TRUE);
+	if (rv != NULL)
+		return rv;
+
+	if (apr_jwk_parse_symmetric_key(cmd->pool, kid,
+			(const unsigned char *) secret, key_len, &jwk, &err) == FALSE) {
+		return apr_psprintf(cmd->pool,
+				"apr_jwk_parse_symmetric_key failed for (kid=%s) \"%s\": %s",
+				kid, secret, apr_jwt_e2s(cmd->pool, err));
+	}
+
+	if (*shared_keys == NULL)
+		*shared_keys = apr_hash_make(cmd->pool);
+	apr_hash_set(*shared_keys, jwk->kid,
+			APR_HASH_KEY_STRING, jwk);
+
 	return NULL;
 }
 
@@ -628,6 +768,44 @@ static const char * oidc_set_pass_idtoken_as(cmd_parms *cmd, void *dummy,
 	return NULL;
 }
 
+static int oidc_oauth_accept_token_as_str2int(oidc_dir_cfg *dir_cfg, const char *v) {
+	if (apr_strnatcmp(v, "header") == 0)
+		return OIDC_OAUTH_ACCEPT_TOKEN_IN_HEADER;
+	if (apr_strnatcmp(v, "post") == 0)
+		return OIDC_OAUTH_ACCEPT_TOKEN_IN_POST;
+	if (apr_strnatcmp(v, "query") == 0)
+		return OIDC_OAUTH_ACCEPT_TOKEN_IN_QUERY;
+	if (strstr(v, "cookie") == v) {
+		const char *p = strstr(v, ":");
+		if (p) {
+			p++;
+		} else {
+			p = "PA.global";
+		}
+		apr_hash_set(dir_cfg->oauth_accept_token_options, "cookie-name", APR_HASH_KEY_STRING, p);
+		return OIDC_OAUTH_ACCEPT_TOKEN_IN_COOKIE;
+	}
+	return -1;
+}
+
+/*
+ * define which method of pass an OAuth Bearer token is accepted
+ */
+static const char * oidc_set_accept_oauth_token_in(cmd_parms *cmd, void *m,
+		const char *arg) {
+	oidc_dir_cfg *dir_cfg = (oidc_dir_cfg *) m;
+
+	int b = oidc_oauth_accept_token_as_str2int(dir_cfg, arg);
+	if (b != -1) {
+		dir_cfg->oauth_accept_token_in |= b;
+	} else {
+		return apr_psprintf(cmd->pool, "Invalid value \"%s\" for directive %s",
+				arg, cmd->directive->directive);
+	}
+
+	return NULL;
+}
+
 /*
  * set the syntax of the token expiry claim in the introspection response
  */
@@ -767,14 +945,12 @@ static const char * oidc_set_unauth_action(cmd_parms *cmd, void *m,
 		return NULL;
 	}
 
-	return "parameter must be one of 'auth', 'pass' or '401";
-}
+	if (apr_strnatcmp(arg, "410") == 0) {
+		dir_cfg->unauth_action = RETURN410;
+		return NULL;
+	}
 
-/*
- * OIDCReturn401 is deprecated
- */
-static const char * oidc_return_401(cmd_parms *cmd, void *struct_ptr, int arg) {
-	return "this primitive is deprecated: use \"OIDCUnAuthAction 401\" instead";
+	return "parameter must be one of 'auth', 'pass', '401' or '410'";
 }
 
 /*
@@ -817,6 +993,7 @@ void *oidc_create_server_config(apr_pool_t *pool, server_rec *svr) {
 	c->provider.idtoken_iat_slack = OIDC_DEFAULT_IDTOKEN_IAT_SLACK;
 	c->provider.session_max_duration = OIDC_DEFAULT_SESSION_MAX_DURATION;
 	c->provider.auth_request_params = NULL;
+	c->provider.pkce_method = NULL;
 
 	c->provider.client_jwks_uri = NULL;
 	c->provider.id_token_signed_response_alg = NULL;
@@ -866,6 +1043,7 @@ void *oidc_create_server_config(apr_pool_t *pool, server_rec *svr) {
 
 	c->metadata_dir = NULL;
 	c->session_type = OIDC_DEFAULT_SESSION_TYPE;
+	c->persistent_session_cookie = 0;
 
 	c->http_timeout_long = OIDC_DEFAULT_HTTP_TIMEOUT_LONG;
 	c->http_timeout_short = OIDC_DEFAULT_HTTP_TIMEOUT_SHORT;
@@ -884,6 +1062,9 @@ void *oidc_create_server_config(apr_pool_t *pool, server_rec *svr) {
 	c->crypto_passphrase = NULL;
 
 	c->scrub_request_headers = OIDC_DEFAULT_SCRUB_REQUEST_HEADERS;
+	c->error_template = NULL;
+
+	c->provider.userinfo_refresh_interval = OIDC_DEFAULT_USERINFO_REFRESH_INTERVAL;
 
 	return c;
 }
@@ -1009,6 +1190,9 @@ void *oidc_merge_server_config(apr_pool_t *pool, void *BASE, void *ADD) {
 			add->provider.auth_request_params != NULL ?
 					add->provider.auth_request_params :
 					base->provider.auth_request_params;
+	c->provider.pkce_method =
+			add->provider.pkce_method != NULL ?
+					add->provider.pkce_method : base->provider.pkce_method;
 
 	c->provider.client_jwks_uri =
 			add->provider.client_jwks_uri != NULL ?
@@ -1148,8 +1332,10 @@ void *oidc_merge_server_config(apr_pool_t *pool, void *BASE, void *ADD) {
 			add->cache_shm_size_max != OIDC_DEFAULT_CACHE_SHM_SIZE ?
 					add->cache_shm_size_max : base->cache_shm_size_max;
 	c->cache_shm_entry_size_max =
-			add->cache_shm_entry_size_max != OIDC_DEFAULT_CACHE_SHM_ENTRY_SIZE_MAX ?
-					add->cache_shm_entry_size_max : base->cache_shm_entry_size_max;
+			add->cache_shm_entry_size_max
+					!= OIDC_DEFAULT_CACHE_SHM_ENTRY_SIZE_MAX ?
+					add->cache_shm_entry_size_max :
+					base->cache_shm_entry_size_max;
 
 #ifdef USE_LIBHIREDIS
 	c->cache_redis_server =
@@ -1165,6 +1351,10 @@ void *oidc_merge_server_config(apr_pool_t *pool, void *BASE, void *ADD) {
 	c->session_type =
 			add->session_type != OIDC_DEFAULT_SESSION_TYPE ?
 					add->session_type : base->session_type;
+	c->persistent_session_cookie =
+			add->persistent_session_cookie != 0 ?
+					add->persistent_session_cookie :
+					base->persistent_session_cookie;
 
 	c->cookie_domain =
 			add->cookie_domain != NULL ?
@@ -1203,6 +1393,16 @@ void *oidc_merge_server_config(apr_pool_t *pool, void *BASE, void *ADD) {
 			add->scrub_request_headers != OIDC_DEFAULT_SCRUB_REQUEST_HEADERS ?
 					add->scrub_request_headers : base->scrub_request_headers;
 
+	c->error_template =
+			add->error_template != NULL ?
+					add->error_template : base->error_template;
+
+	c->provider.userinfo_refresh_interval =
+			add->provider.userinfo_refresh_interval
+				!= OIDC_DEFAULT_USERINFO_REFRESH_INTERVAL ?
+					add->provider.userinfo_refresh_interval :
+					base->provider.userinfo_refresh_interval;
+
 	return c;
 }
 
@@ -1219,6 +1419,9 @@ void *oidc_create_dir_config(apr_pool_t *pool, char *path) {
 	c->pass_cookies = apr_array_make(pool, 0, sizeof(const char *));
 	c->pass_info_in_headers = 1;
 	c->pass_info_in_env_vars = 1;
+	c->oauth_accept_token_in = OIDC_OAUTH_ACCEPT_TOKEN_IN_DEFAULT;
+	c->oauth_accept_token_options = apr_hash_make(pool);
+	c->oauth_token_introspect_interval = 0;
 	return (c);
 }
 
@@ -1252,6 +1455,16 @@ void *oidc_merge_dir_config(apr_pool_t *pool, void *BASE, void *ADD) {
 	c->pass_info_in_env_vars = (
 			add->pass_info_in_env_vars != 1 ?
 					add->pass_info_in_env_vars : base->pass_info_in_env_vars);
+	c->oauth_accept_token_in = (
+			add->oauth_accept_token_in != OIDC_OAUTH_ACCEPT_TOKEN_IN_DEFAULT ?
+					add->oauth_accept_token_in : base->oauth_accept_token_in);
+	c->oauth_accept_token_options = apr_hash_merge(pool,
+			add->oauth_accept_token_options, base->oauth_accept_token_options,
+			NULL, NULL);
+	c->oauth_token_introspect_interval = (
+			add->oauth_token_introspect_interval != 0 ?
+					add->oauth_token_introspect_interval :
+					base->oauth_token_introspect_interval);
 	return (c);
 }
 
@@ -1289,9 +1502,6 @@ static int oidc_check_config_openid_openidc(server_rec *s, oidc_cfg *c) {
 			if (c->provider.authorization_endpoint_url == NULL)
 				return oidc_check_config_error(s,
 						"OIDCProviderAuthorizationEndpoint");
-			// TODO: this depends on the configured OIDCResponseType now
-			//			if (c->provider.token_endpoint_url == NULL)
-			//				return oidc_check_config_error(s, "OIDCProviderTokenEndpoint");
 		} else {
 			apr_uri_parse(s->process->pconf, c->provider.metadata_url, &r_uri);
 			if (apr_strnatcmp(r_uri.scheme, "http") == 0) {
@@ -1302,9 +1512,6 @@ static int oidc_check_config_openid_openidc(server_rec *s, oidc_cfg *c) {
 		}
 		if (c->provider.client_id == NULL)
 			return oidc_check_config_error(s, "OIDCClientID");
-		// TODO: this depends on the configured OIDCResponseType now
-		if (c->provider.client_secret == NULL)
-			return oidc_check_config_error(s, "OIDCClientSecret");
 	} else {
 		if (c->provider.metadata_url != NULL) {
 			oidc_serror(s,
@@ -1321,8 +1528,7 @@ static int oidc_check_config_openid_openidc(server_rec *s, oidc_cfg *c) {
 	}
 
 	if (c->cookie_domain != NULL) {
-		char *p = strstr(r_uri.hostname, c->cookie_domain);
-		if ((p == NULL) || (apr_strnatcmp(c->cookie_domain, p) != 0)) {
+		if (!oidc_util_cookie_domain_valid(r_uri.hostname, c->cookie_domain)) {
 			oidc_serror(s,
 					"the domain (%s) configured in OIDCCookieDomain does not match the URL hostname (%s) of the configured OIDCRedirectURI (%s): setting \"state\" and \"session\" cookies will not work!",
 					c->cookie_domain, r_uri.hostname, c->redirect_uri);
@@ -1508,7 +1714,19 @@ static int oidc_post_config(apr_pool_t *pool, apr_pool_t *p1, apr_pool_t *p2,
 		return OK;
 	}
 
-	ap_log_error(APLOG_MARK, APLOG_INFO, 0, s, "%s - init", NAMEVERSION);
+	ap_log_error(APLOG_MARK, APLOG_INFO, 0, s,
+			"%s - init - %s, Elliptic Curve (EC)=%s, Galois Counter Mode (GCM)=%s, Redis=%s",
+			NAMEVERSION,
+			OPENSSL_VERSION_TEXT,
+			APR_JWS_EC_SUPPORT  ? "yes" : "no",
+			APR_JWE_GCM_SUPPORT ? "yes" : "no",
+#ifdef USE_LIBHIREDIS
+			"yes"
+#else
+			"no"
+#endif
+		);
+
 
 	curl_global_init(CURL_GLOBAL_ALL);
 	OpenSSL_add_all_digests();
@@ -1781,6 +1999,11 @@ const command_rec oidc_config_cmds[] = {
 				(void*)APR_OFFSETOF(oidc_cfg, provider.auth_request_params),
 				RSRC_CONF,
 				"Extra parameters that need to be sent in the Authorization Request (must be query-encoded like \"display=popup&prompt=consent\"."),
+		AP_INIT_TAKE1("OIDCPKCEMethod",
+				oidc_set_pkce_method,
+				(void *)APR_OFFSETOF(oidc_cfg, provider.pkce_method),
+				RSRC_CONF,
+				"The RFC 7636 PCKE mode used; must be one of \"plain\", \"S256\""),
 
 		AP_INIT_TAKE1("OIDCClientID", oidc_set_string_slot,
 				(void*)APR_OFFSETOF(oidc_cfg, provider.client_id),
@@ -1799,7 +2022,8 @@ const command_rec oidc_config_cmds[] = {
 				(void *)APR_OFFSETOF(oidc_cfg, default_sso_url),
 				RSRC_CONF,
 				"Defines the default URL where the user is directed to in case of 3rd-party initiated SSO."),
-		AP_INIT_TAKE1("OIDCDefaultLoggedOutURL", oidc_set_url_slot,
+		AP_INIT_TAKE1("OIDCDefaultLoggedOutURL",
+				oidc_set_url_slot,
 				(void *)APR_OFFSETOF(oidc_cfg, default_slo_url),
 				RSRC_CONF,
 				"Defines the default URL where the user is directed to after logout."),
@@ -1931,7 +2155,7 @@ const command_rec oidc_config_cmds[] = {
 		AP_INIT_TAKE1("OIDCSessionType", oidc_set_session_type,
 				(void*)APR_OFFSETOF(oidc_cfg, session_type),
 				RSRC_CONF,
-				"OpenID Connect session storage type (Apache 2.0/2.2 only). Must be one of \"server-cache\" or \"client-cookie\"."),
+				"OpenID Connect session storage type (Apache 2.0/2.2 only). Must be one of \"server-cache\" or \"client-cookie\" with an optional suffix \":persistent\"."),
 		AP_INIT_FLAG("OIDCScrubRequestHeaders",
 				oidc_set_flag_slot,
 				(void *) APR_OFFSETOF(oidc_cfg, scrub_request_headers),
@@ -1960,7 +2184,8 @@ const command_rec oidc_config_cmds[] = {
 				(void*)APR_OFFSETOF(oidc_cfg, cache_shm_size_max),
 				RSRC_CONF,
 				"Maximum number of cache entries to use for \"shm\" caching."),
-		AP_INIT_TAKE1("OIDCCacheShmEntrySizeMax", oidc_set_cache_shm_entry_size_max,
+		AP_INIT_TAKE1("OIDCCacheShmEntrySizeMax",
+				oidc_set_cache_shm_entry_size_max,
 				(void*)APR_OFFSETOF(oidc_cfg, cache_shm_entry_size_max),
 				RSRC_CONF,
 				"Maximum size of a single cache entry used for \"shm\" caching."),
@@ -1976,6 +2201,11 @@ const command_rec oidc_config_cmds[] = {
 				RSRC_CONF,
 				"Password for authentication to the Redis servers."),
 #endif
+		AP_INIT_TAKE1("OIDCHTMLErrorTemplate",
+				oidc_set_string_slot,
+				(void*)APR_OFFSETOF(oidc_cfg, error_template),
+				RSRC_CONF,
+				"Name of a HTML error template: needs to contain two \"%s\" characters, one for the error message, one for the description."),
 
 		AP_INIT_TAKE1("OIDCDiscoverURL", oidc_set_url_slot_dir_cfg,
 				(void *)APR_OFFSETOF(oidc_dir_cfg, discover_url),
@@ -1998,17 +2228,29 @@ const command_rec oidc_config_cmds[] = {
 				(void *) APR_OFFSETOF(oidc_dir_cfg, cookie),
 				RSRC_CONF|ACCESS_CONF|OR_AUTHCFG,
 				"Define the cookie name for the session cookie."),
-		AP_INIT_FLAG("OIDCReturn401", oidc_return_401,
+		AP_INIT_TAKE1("OIDCUnAuthAction",
+				oidc_set_unauth_action,
 				(void *) APR_OFFSETOF(oidc_dir_cfg, unauth_action),
 				RSRC_CONF|ACCESS_CONF|OR_AUTHCFG,
-				"Indicates whether a user will be redirected to the Provider when not authenticated (Off) or a 401 will be returned (On)."),
-		AP_INIT_TAKE1("OIDCUnAuthAction", oidc_set_unauth_action,
-				(void *) APR_OFFSETOF(oidc_dir_cfg, unauth_action),
-				RSRC_CONF|ACCESS_CONF|OR_AUTHCFG,
-				"Sets the action taken when an unauthenticated request occurs: \"auth\" (default) means redirect to OP for authentication, \"pass\" means always allow the request but pass claims when authenticated, \"401\" means return HTTP 401."),
+				"Sets the action taken when an unauthenticated request occurs: must be one of \"auth\" (default), \"pass\" , \"401\" or \"410\"."),
 		AP_INIT_TAKE1("OIDCPassClaimsAs",
 				oidc_set_pass_claims_as, NULL,
 				RSRC_CONF|ACCESS_CONF|OR_AUTHCFG,
 				"Specify how claims are passed to the application(s); must be one of \"none\", \"headers\", \"environment\" or \"both\" (default)."),
+		AP_INIT_ITERATE("OIDCOAuthAcceptTokenAs",
+				oidc_set_accept_oauth_token_in,
+				NULL,
+				RSRC_CONF|ACCESS_CONF|OR_AUTHCFG,
+				"The method in which an OAuth token can be presented; must be one or more of: header|post|query|cookie"),
+		AP_INIT_TAKE1("OIDCUserInfoRefreshInterval",
+				oidc_set_int_slot,
+				(void*)APR_OFFSETOF(oidc_cfg, provider.userinfo_refresh_interval),
+				RSRC_CONF,
+				"Duration in seconds after which retrieved claims from the userinfo endpoint should be refreshed."),
+		AP_INIT_TAKE1("OIDCOAuthTokenIntrospectionInterval",
+				ap_set_int_slot,
+				(void *) APR_OFFSETOF(oidc_dir_cfg, oauth_token_introspect_interval),
+				RSRC_CONF|ACCESS_CONF|OR_AUTHCFG,
+				"Sets the token introspection refresh interval."),
 		{ NULL }
 };
