@@ -18,7 +18,7 @@
  */
 
 /***************************************************************************
- * Copyright (C) 2013-2015 Ping Identity Corporation
+ * Copyright (C) 2013-2016 Ping Identity Corporation
  * All rights reserved.
  *
  * For further information please contact:
@@ -89,7 +89,7 @@ apr_array_header_t *apr_jwe_supported_encryptions(apr_pool_t *pool) {
 	*(const char**) apr_array_push(result) = "A128CBC-HS256";
 	*(const char**) apr_array_push(result) = "A192CBC-HS384";
 	*(const char**) apr_array_push(result) = "A256CBC-HS512";
-#if (OPENSSL_VERSION_NUMBER >= 0x1000100f)
+#if (APR_JWE_GCM_SUPPORT)
 	*(const char**) apr_array_push(result) = "A128GCM";
 	*(const char**) apr_array_push(result) = "A192GCM";
 	*(const char**) apr_array_push(result) = "A256GCM";
@@ -125,7 +125,7 @@ static const EVP_CIPHER *apr_jwe_enc_to_openssl_cipher(const char *enc) {
 	if (apr_strnatcmp(enc, "A256CBC-HS512") == 0) {
 		return EVP_aes_256_cbc();
 	}
-#if (OPENSSL_VERSION_NUMBER >= 0x1000100f)
+#if (APR_JWE_GCM_SUPPORT)
 	if (apr_strnatcmp(enc, "A128CM") == 0) {
 		return EVP_aes_128_gcm();
 	}
@@ -371,7 +371,7 @@ static apr_byte_t apr_jwe_decrypt_cek(apr_pool_t *pool,
 	return rc;
 }
 
-#if (OPENSSL_VERSION_NUMBER >= 0x1000100f)
+#if (APR_JWE_GCM_SUPPORT)
 /*
  * Decrypt AES-GCM content
  */
@@ -382,7 +382,7 @@ apr_byte_t apr_jwe_decrypt_content_aesgcm(apr_pool_t *pool,
 		apr_jwt_error_t *err) {
 
 	EVP_CIPHER_CTX *ctx;
-	int outlen, rv;
+	int rv;
 
 	ctx = EVP_CIPHER_CTX_new();
 	if (!EVP_DecryptInit_ex(ctx, apr_jwe_enc_to_openssl_cipher(header->enc),
@@ -391,8 +391,9 @@ apr_byte_t apr_jwe_decrypt_content_aesgcm(apr_pool_t *pool,
 		return FALSE;
 	}
 
+	int p_len = cipher_text->len, f_len = 0;
 	unsigned char *plaintext = apr_palloc(pool,
-			cipher_text->len
+			p_len
 			+ EVP_CIPHER_block_size(
 					apr_jwe_enc_to_openssl_cipher(header->enc)));
 
@@ -407,13 +408,13 @@ apr_byte_t apr_jwe_decrypt_content_aesgcm(apr_pool_t *pool,
 		return FALSE;
 	}
 	/* zero or more calls to specify any AAD */
-	if (!EVP_DecryptUpdate(ctx, NULL, &outlen, (unsigned char *) aad,
+	if (!EVP_DecryptUpdate(ctx, NULL, &p_len, (unsigned char *) aad,
 			aad_len)) {
 		apr_jwt_error_openssl(err, "EVP_DecryptUpdate (aad)");
 		return FALSE;
 	}
 	/* decrypt plaintext */
-	if (!EVP_DecryptUpdate(ctx, plaintext, &outlen,
+	if (!EVP_DecryptUpdate(ctx, plaintext, &p_len,
 			(unsigned char *) cipher_text->value, cipher_text->len)) {
 		apr_jwt_error_openssl(err, "EVP_DecryptUpdate (ciphertext)");
 		return FALSE;
@@ -425,7 +426,8 @@ apr_byte_t apr_jwe_decrypt_content_aesgcm(apr_pool_t *pool,
 	}
 
 	/* finalise: note get no output for GCM */
-	rv = EVP_DecryptFinal_ex(ctx, plaintext, &outlen);
+	rv = EVP_DecryptFinal_ex(ctx, plaintext, &f_len);
+	plaintext[p_len + f_len] = '\0';
 
 	EVP_CIPHER_CTX_free(ctx);
 
@@ -606,7 +608,7 @@ apr_byte_t apr_jwe_decrypt_jwt(apr_pool_t *pool, apr_jwt_header_t *header,
 				cipher_text, cek, cek_len, iv, aad, aad_len, auth_tag,
 				decrypted, err_r);
 
-#if (OPENSSL_VERSION_NUMBER >= 0x1000100f)
+#if (APR_JWE_GCM_SUPPORT)
 
 	} else if ((apr_strnatcmp(header->enc, "A128GCM") == 0)
 			|| (apr_strnatcmp(header->enc, "A192GCM") == 0)
