@@ -18,7 +18,7 @@
  */
 
 /***************************************************************************
- * Copyright (C) 2013-2015 Ping Identity Corporation
+ * Copyright (C) 2013-2016 Ping Identity Corporation
  * All rights reserved.
  *
  * For further information please contact:
@@ -57,6 +57,8 @@
 #include <openssl/hmac.h>
 #include <openssl/err.h>
 
+#include <openssl/ecdsa.h>
+
 #include <apr_base64.h>
 
 #include "apr_jose.h"
@@ -78,7 +80,7 @@ apr_array_header_t *apr_jws_supported_algorithms(apr_pool_t *pool) {
 	*(const char**) apr_array_push(result) = "HS256";
 	*(const char**) apr_array_push(result) = "HS384";
 	*(const char**) apr_array_push(result) = "HS512";
-#if (OPENSSL_VERSION_NUMBER >= 0x01000000)
+#if (APR_JWS_EC_SUPPORT)
 	*(const char**) apr_array_push(result) = "ES256";
 	*(const char**) apr_array_push(result) = "ES384";
 	*(const char**) apr_array_push(result) = "ES512";
@@ -153,17 +155,9 @@ const EVP_MD *apr_jws_crypto_alg_to_evp(apr_pool_t *pool, const char *alg,
 }
 
 /*
- * verify HMAC signature on JWT
+ * calculate JWT HMAC signature
  */
-static apr_byte_t apr_jws_verify_hmac(apr_pool_t *pool, apr_jwt_t *jwt,
-		apr_jwk_t *jwk, apr_jwt_error_t *err) {
-
-	if (jwk->type != APR_JWK_KEY_OCT) {
-		apr_jwt_error(err,
-				"key type of provided JWK cannot be used for HMAC verification: %d",
-				jwk->type);
-		return FALSE;
-	}
+apr_byte_t apr_jws_calculate_hmac(apr_pool_t *pool, apr_jwt_t *jwt, apr_jwk_t *jwk, unsigned char *md, unsigned int *md_len, apr_jwt_error_t *err) {
 
 	/* get the OpenSSL digest function */
 	const EVP_MD *digest = NULL;
@@ -174,14 +168,28 @@ static apr_byte_t apr_jws_verify_hmac(apr_pool_t *pool, apr_jwt_t *jwt,
 	unsigned char *msg = (unsigned char *) jwt->message;
 	unsigned int msg_len = strlen(jwt->message);
 
+	/* apply the HMAC function to the message with the provided key */
+	if (!HMAC(digest, jwk->key.oct->k, jwk->key.oct->k_len, msg, msg_len, md,
+			md_len)) {
+		apr_jwt_error_openssl(err, "HMAC");
+		return FALSE;
+	}
+
+	return TRUE;
+}
+
+/*
+ * verify HMAC signature on JWT
+ */
+static apr_byte_t apr_jws_verify_hmac(apr_pool_t *pool, apr_jwt_t *jwt,
+		apr_jwk_t *jwk, apr_jwt_error_t *err) {
+
 	/* prepare the hash */
 	unsigned int md_len = 0;
 	unsigned char md[EVP_MAX_MD_SIZE];
 
-	/* apply the HMAC function to the message with the provided key */
-	if (!HMAC(digest, jwk->key.oct->k, jwk->key.oct->k_len, msg, msg_len, md,
-			&md_len)) {
-		apr_jwt_error_openssl(err, "HMAC");
+	/* calculate the HMAC hash */
+	if (apr_jws_calculate_hmac(pool, jwt, jwk, (unsigned char *)&md, &md_len, err) == FALSE) {
 		return FALSE;
 	}
 
@@ -281,6 +289,120 @@ int apr_jws_hash_length(const char *alg) {
 	return 0;
 }
 
+/*
+ * calculate JWT RSA signature
+ */
+apr_byte_t apr_jws_calculate_rsa(apr_pool_t *pool, apr_jwt_t *jwt,
+		apr_jwk_t *jwk, unsigned char *md, unsigned int *md_len,
+		apr_jwt_error_t *err) {
+
+	apr_byte_t rc = FALSE;
+
+	/* get the OpenSSL digest function */
+	const EVP_MD *digest = NULL;
+	if ((digest = apr_jws_crypto_alg_to_evp(pool, jwt->header.alg, err)) == NULL)
+		return FALSE;
+
+	EVP_MD_CTX ctx;
+	EVP_MD_CTX_init(&ctx);
+
+	RSA * privkey = RSA_new();
+
+	BIGNUM * modulus = BN_new();
+	BIGNUM * exponent = BN_new();
+	BIGNUM * private_exponent = BN_new();
+
+	BN_bin2bn(jwk->key.rsa->modulus, jwk->key.rsa->modulus_len, modulus);
+	BN_bin2bn(jwk->key.rsa->exponent, jwk->key.rsa->exponent_len, exponent);
+	BN_bin2bn(jwk->key.rsa->private_exponent,
+			jwk->key.rsa->private_exponent_len, private_exponent);
+
+	privkey->n = modulus;
+	privkey->e = exponent;
+	privkey->d = private_exponent;
+
+	EVP_PKEY* pRsaKey = EVP_PKEY_new();
+	if (!EVP_PKEY_assign_RSA(pRsaKey, privkey)) {
+		pRsaKey = NULL;
+		apr_jwt_error_openssl(err, "EVP_PKEY_assign_RSA");
+		goto end;
+	}
+
+	if (apr_jws_signature_starts_with(pool, jwt->header.alg, "PS") == TRUE) {
+
+		unsigned char *pDigest = apr_pcalloc(pool, RSA_size(privkey));
+		unsigned int uDigestLen = RSA_size(privkey);
+
+		if (!EVP_DigestInit(&ctx, digest)) {
+			apr_jwt_error_openssl(err, "EVP_DigestInit");
+			goto end;
+		}
+		if (!EVP_DigestUpdate(&ctx, jwt->message, strlen(jwt->message))) {
+			apr_jwt_error_openssl(err, "EVP_DigestUpdate");
+			goto end;
+		}
+		if (!EVP_DigestFinal(&ctx, pDigest, &uDigestLen)) {
+			apr_jwt_error_openssl(err, "wrong key? EVP_DigestFinal");
+			goto end;
+		}
+
+		unsigned char *pPadded = apr_pcalloc(pool, RSA_size(privkey));
+
+		int status = 0;
+		status = RSA_padding_add_PKCS1_PSS(privkey, pPadded, pDigest, digest,
+				-2 /* maximum salt length*/);
+		if (!status) {
+			apr_jwt_error_openssl(err, "RSA_padding_add_PKCS1_PSS");
+			goto end;
+		}
+
+		jwt->signature.length = RSA_size(privkey);
+		status = RSA_private_encrypt(jwt->signature.length, pPadded,
+				jwt->signature.bytes, privkey, RSA_NO_PADDING);
+		if (status == -1) {
+			apr_jwt_error_openssl(err,
+					apr_psprintf(pool,
+							"RSA_private_encrypt: digest_len=%d, sig_len=%d",
+							uDigestLen, jwt->signature.length));
+			goto end;
+		}
+
+		rc = TRUE;
+
+	} else {
+
+		if (!EVP_SignInit_ex(&ctx, digest, NULL)) {
+			apr_jwt_error_openssl(err, "EVP_SignInit_ex");
+			goto end;
+		}
+
+		if (!EVP_SignUpdate(&ctx, jwt->message, strlen(jwt->message))) {
+			apr_jwt_error_openssl(err, "EVP_SignUpdate");
+			goto end;
+		}
+
+		if (!EVP_SignFinal(&ctx, (unsigned char *) jwt->signature.bytes,
+				(unsigned int *) &jwt->signature.length, pRsaKey)) {
+			apr_jwt_error_openssl(err, "wrong key? EVP_SignFinal");
+			goto end;
+		}
+
+		rc = TRUE;
+
+	}
+
+	end:
+
+	if (pRsaKey) {
+		EVP_PKEY_free(pRsaKey);
+	} else if (privkey) {
+		RSA_free(privkey);
+	}
+	EVP_MD_CTX_cleanup(&ctx);
+
+	return rc;
+}
+
 /*
  * verify HMAC signature on JWT
  */
@@ -363,9 +485,15 @@ static apr_byte_t apr_jws_verify_rsa(apr_pool_t *pool, apr_jwt_t *jwt,
 			apr_jwt_error_openssl(err, "EVP_VerifyUpdate");
 			goto end;
 		}
-		if (!EVP_VerifyFinal(&ctx, (const unsigned char *) jwt->signature.bytes,
-				jwt->signature.length, pRsaKey)) {
-			apr_jwt_error_openssl(err, "wrong key? EVP_VerifyFinal");
+		
+		int rv = EVP_VerifyFinal(&ctx, (const unsigned char *) jwt->signature.bytes,
+				jwt->signature.length, pRsaKey);
+
+		if (rv < 0) {
+			apr_jwt_error_openssl(err, "EVP_VerifyFinal");
+			goto end;
+		} else if (rv == 0) {
+			apr_jwt_error(err, "wrong key");
 			goto end;
 		}
 
@@ -385,7 +513,7 @@ end:
 	return rc;
 }
 
-#if (OPENSSL_VERSION_NUMBER >= 0x01000000)
+#if (APR_JWS_EC_SUPPORT)
 
 /*
  * return the OpenSSL Elliptic Curve NID for a JWT algorithm
@@ -425,12 +553,8 @@ static apr_byte_t apr_jws_verify_ec(apr_pool_t *pool, apr_jwt_t *jwt,
 	apr_byte_t rc = FALSE;
 
 	/* get the OpenSSL digest function */
-	const EVP_MD *digest = NULL;
-	if ((digest = apr_jws_crypto_alg_to_evp(pool, jwt->header.alg, err)) == NULL)
-		return FALSE;
-
-	EVP_MD_CTX ctx;
-	EVP_MD_CTX_init(&ctx);
+	const char *digest = apr_jws_alg_to_openssl_digest(jwt->header.alg);
+	if (digest == NULL) return FALSE;
 
 	EC_KEY * pubkey = EC_KEY_new();
 	EC_KEY_set_group(pubkey, curve);
@@ -446,43 +570,50 @@ static apr_byte_t apr_jws_verify_ec(apr_pool_t *pool, apr_jwt_t *jwt,
 		return FALSE;
 	}
 
-	EVP_PKEY* pEcKey = EVP_PKEY_new();
-	if (!EVP_PKEY_assign_EC_KEY(pEcKey, pubkey)) {
-		pEcKey = NULL;
-		apr_jwt_error_openssl(err, "EVP_PKEY_assign_EC_KEY");
+	//P-256:  64 byte signature
+	//P-384:  96 byte signature
+	//P-521: 132 byte signature
+	int key_len = jwt->signature.length / 2;
+
+	ECDSA_SIG *ecdsa_sig = NULL;
+	ecdsa_sig = ECDSA_SIG_new();
+
+#if OPENSSL_VERSION_NUMBER >= 0x10100005L
+	BIGNUM *r = BN_new();
+	BIGNUM *s = BN_new();
+	BN_bin2bn(jwt->signature.bytes, key_len, r);
+	BN_bin2bn(jwt->signature.bytes + key_len, key_len, s);
+	ECDSA_SIG_set0(ecdsa_sig, r, s);
+#else
+	BN_bin2bn(jwt->signature.bytes, key_len, ecdsa_sig->r);
+	BN_bin2bn(jwt->signature.bytes + key_len, key_len, ecdsa_sig->s);
+#endif
+
+	char *hash = NULL;
+	int hash_len = 0;
+	if (apr_jws_hash_bytes(pool, digest, (const unsigned char *)jwt->message, (unsigned int)strlen(jwt->message), (unsigned char **)&hash, (unsigned int *)&hash_len, err) == FALSE) {
+ 		apr_jwt_error(err, "apr_jws_hash_bytes");
 		goto end;
 	}
 
-	ctx.pctx = EVP_PKEY_CTX_new(pEcKey, NULL);
-
-	if (!EVP_PKEY_verify_init(ctx.pctx)) {
-		apr_jwt_error_openssl(err, "EVP_PKEY_verify_init");
+    int rv = ECDSA_do_verify((const unsigned char *)hash, hash_len, ecdsa_sig, pubkey);
+ 	if (rv < 0) {
+ 		apr_jwt_error_openssl(err, "ECDSA_do_verify");
 		goto end;
-	}
-	if (!EVP_VerifyInit_ex(&ctx, digest, NULL)) {
-		apr_jwt_error_openssl(err, "EVP_VerifyInit_ex");
+ 	} else if (rv == 0) {
+ 		apr_jwt_error(err, "wrong key");
 		goto end;
-	}
-	if (!EVP_VerifyUpdate(&ctx, jwt->message, strlen(jwt->message))) {
-		apr_jwt_error_openssl(err, "EVP_VerifyUpdate");
-		goto end;
-	}
-	if (!EVP_VerifyFinal(&ctx, (const unsigned char *) jwt->signature.bytes,
-			jwt->signature.length, pEcKey)) {
-		apr_jwt_error_openssl(err, "wrong key? EVP_VerifyFinal");
-		goto end;
-	}
+ 	}
 
 	rc = TRUE;
 
 end:
 
-	if (pEcKey) {
-		EVP_PKEY_free(pEcKey);
-	} else if (pubkey) {
+	if (ecdsa_sig)
+		ECDSA_SIG_free(ecdsa_sig);
+	if (pubkey) {
 		EC_KEY_free(pubkey);
 	}
-	EVP_MD_CTX_cleanup(&ctx);
 
 	return rc;
 }
@@ -504,7 +635,7 @@ apr_byte_t apr_jws_signature_is_rsa(apr_pool_t *pool, apr_jwt_t *jwt) {
 			|| apr_jws_signature_starts_with(pool, jwt->header.alg, "PS");
 }
 
-#if (OPENSSL_VERSION_NUMBER >= 0x01000000)
+#if (APR_JWS_EC_SUPPORT)
 
 /*
  * check if the signature on the JWT is Elliptic Curve based
@@ -533,7 +664,7 @@ static apr_byte_t apr_jws_verify_with_jwk(apr_pool_t *pool, apr_jwt_t *jwt,
 		rc = (jwk->type == APR_JWK_KEY_RSA)
 				&& apr_jws_verify_rsa(pool, jwt, jwk, err);
 
-#if (OPENSSL_VERSION_NUMBER >= 0x01000000)
+#if (APR_JWS_EC_SUPPORT)
 	} else if (apr_jws_signature_is_ec(pool, jwt)) {
 
 		rc = (jwk->type == APR_JWK_KEY_EC)
