@@ -18,7 +18,7 @@
  */
 
 /***************************************************************************
- * Copyright (C) 2013-2015 Ping Identity Corporation
+ * Copyright (C) 2013-2016 Ping Identity Corporation
  * All rights reserved.
  *
  * For further information please contact:
@@ -160,7 +160,7 @@ static apr_byte_t oidc_metadata_file_read_json(request_rec *r, const char *path,
 	char *buf = NULL;
 
 	/* read the file contents */
-	if (oidc_util_file_read(r, path, &buf) == FALSE)
+	if (oidc_util_file_read(r, path, r->pool, &buf) == FALSE)
 		return FALSE;
 
 	/* decode the JSON contents of the buffer */
@@ -241,10 +241,10 @@ static apr_byte_t oidc_metadata_provider_is_valid(request_rec *r,
 	/* check that the issuer matches */
 	if (issuer != NULL) {
 		if (oidc_util_issuer_match(issuer, json_string_value(j_issuer)) == FALSE) {
-			oidc_warn(r,
+			oidc_error(r,
 					"requested issuer (%s) does not match the \"issuer\" value in the provider metadata file: %s",
 					issuer, json_string_value(j_issuer));
-			//return FALSE;
+			return FALSE;
 		}
 	}
 
@@ -356,10 +356,16 @@ static apr_byte_t oidc_metadata_provider_is_valid(request_rec *r,
 			if (strcmp(json_string_value(elem), "client_secret_basic") == 0) {
 				break;
 			}
+			if (strcmp(json_string_value(elem), "client_secret_jwt") == 0) {
+				break;
+			}
+			if (strcmp(json_string_value(elem), "private_key_jwt") == 0) {
+				break;
+			}
 		}
 		if (i == json_array_size(j_token_endpoint_auth_methods_supported)) {
 			oidc_error(r,
-					"could not find a supported value [client_secret_post|client_secret_basic] in provider (%s) metadata for entry \"token_endpoint_auth_methods_supported\"",
+					"could not find a supported value [client_secret_post|client_secret_basic|client_secret_jwt|private_key_jwt] in provider (%s) metadata for entry \"token_endpoint_auth_methods_supported\"",
 					issuer);
 			return FALSE;
 		}
@@ -580,6 +586,10 @@ static apr_byte_t oidc_metadata_client_register(request_rec *r, oidc_cfg *cfg,
 	}
 	json_object_set_new(data, "response_types", response_types);
 
+	if (provider->token_endpoint_auth != NULL) {
+		json_object_set_new(data, "token_endpoint_auth_method", json_string(provider->token_endpoint_auth));
+	}
+
 	if (provider->client_contact != NULL) {
 		json_object_set_new(data, "contacts",
 				json_pack("[s]", provider->client_contact));
@@ -923,7 +933,7 @@ apr_byte_t oidc_metadata_list(request_rec *r, oidc_cfg *cfg,
 	}
 
 	/* allocate some space in the array that will hold the list of providers */
-	*list = apr_array_make(r->pool, 5, sizeof(sizeof(const char*)));
+	*list = apr_array_make(r->pool, 5, sizeof(const char*));
 	/* BTW: we could estimate the number in the array based on # directory entries... */
 
 	/* loop over the entries in the provider metadata directory */
@@ -1035,11 +1045,14 @@ apr_byte_t oidc_metadata_provider_parse(request_rec *r, json_t *j_provider,
 					continue;
 				}
 
-				/* take first supported method and prefer post over basic */
+				/* take first supported method and prefer post over basic over jwt */
 				if ((apr_strnatcmp(json_string_value(elem),
 						"client_secret_post") == 0)
 						|| (apr_strnatcmp(json_string_value(elem),
-								"client_secret_basic") == 0)) {
+								"client_secret_basic") == 0)
+								|| (apr_strnatcmp(json_string_value(elem),
+										"client_secret_jwt") == 0) || (apr_strnatcmp(json_string_value(elem),
+												"private_key_jwt") == 0)) {
 					token_endpoint_auth = json_string_value(elem);
 					break;
 				}
@@ -1127,6 +1140,10 @@ apr_byte_t oidc_metadata_conf_parse(request_rec *r, oidc_cfg *cfg,
 	oidc_json_object_get_string(r->pool, j_conf, "response_mode",
 			&provider->response_mode, cfg->provider.response_mode);
 
+	/* get the PKCE method to use */
+	oidc_json_object_get_string(r->pool, j_conf, "pkce_method",
+			&provider->pkce_method, cfg->provider.pkce_method);
+
 	/* get the client name */
 	oidc_json_object_get_string(r->pool, j_conf, "client_name",
 			&provider->client_name, cfg->provider.client_name);
@@ -1148,6 +1165,11 @@ apr_byte_t oidc_metadata_conf_parse(request_rec *r, oidc_cfg *cfg,
 	oidc_json_object_get_string(r->pool, j_conf, "response_type",
 			&provider->response_type, NULL);
 
+	/* see if we've got a custom user info refresh interval */
+	oidc_json_object_get_int(r->pool, j_conf, "userinfo_refresh_interval",
+			&provider->userinfo_refresh_interval,
+			cfg->provider.userinfo_refresh_interval);
+
 	return TRUE;
 }
 
@@ -1173,7 +1195,9 @@ apr_byte_t oidc_metadata_client_parse(request_rec *r, oidc_cfg *cfg,
 	if (token_endpoint_auth != NULL) {
 		if ((apr_strnatcmp(token_endpoint_auth, "client_secret_post") == 0)
 				|| (apr_strnatcmp(token_endpoint_auth, "client_secret_basic")
-						== 0)) {
+						== 0)
+						|| (apr_strnatcmp(token_endpoint_auth, "client_secret_jwt") == 0)
+						|| (apr_strnatcmp(token_endpoint_auth, "private_key_jwt") == 0)) {
 			provider->token_endpoint_auth = apr_pstrdup(r->pool,
 					token_endpoint_auth);
 		} else {
