@@ -614,19 +614,14 @@ const char*oidc_request_state_get(request_rec *r, const char *key) {
  * in the session in to HTTP headers passed on to the application
  */
 static apr_byte_t oidc_set_app_claims(request_rec *r,
-		const oidc_cfg * const cfg, session_rec *session,
-		const char *session_key) {
+		const oidc_cfg * const cfg, session_rec *session, const char *s_claims) {
 
 	/* get a handle to the directory config */
 	oidc_dir_cfg *dir_cfg = ap_get_module_config(r->per_dir_config,
 			&auth_openidc_module);
 
-	const char *s_claims = NULL;
 	json_t *j_claims = NULL;
 
-	/* get the string-encoded JSON object from the session */
-	oidc_session_get(r, session, session_key, &s_claims);
-
 	/* decode the string-encoded attributes in to a JSON structure */
 	if (s_claims != NULL) {
 		json_error_t json_error;
@@ -635,8 +630,8 @@ static apr_byte_t oidc_set_app_claims(request_rec *r,
 		if (j_claims == NULL) {
 			/* whoops, JSON has been corrupted */
 			oidc_error(r,
-					"unable to parse \"%s\" JSON stored in the session (%s), returning internal server error",
-					json_error.text, session_key);
+					"unable to parse \"%s\" JSON stored in the session: %s",
+					s_claims, json_error.text);
 
 			return FALSE;
 		}
@@ -648,9 +643,6 @@ static apr_byte_t oidc_set_app_claims(request_rec *r,
 				cfg->claim_delimiter, dir_cfg->pass_info_in_headers,
 				dir_cfg->pass_info_in_env_vars);
 
-		/* set the claims JSON string in the request state so it is available for authz purposes later on */
-		oidc_request_state_set(r, session_key, s_claims);
-
 		/* release resources */
 		json_decref(j_claims);
 	}
@@ -930,11 +922,13 @@ static apr_byte_t oidc_refresh_claims_from_userinfo_endpoint(request_rec *r,
 	if ((provider->userinfo_endpoint_url != NULL) && (interval > 0)) {
 
 		/* get the last refresh timestamp from the session info */
-		apr_time_t last_refresh;
+		apr_time_t last_refresh = 0;
 		const char *s_last_refresh = NULL;
 		oidc_session_get(r, session, OIDC_USERINFO_LAST_REFRESH_SESSION_KEY,
 				&s_last_refresh);
-		sscanf(s_last_refresh, "%" APR_TIME_T_FMT, &last_refresh);
+		if (s_last_refresh != NULL) {
+			sscanf(s_last_refresh, "%" APR_TIME_T_FMT, &last_refresh);
+		}
 
 		oidc_debug(r, "refresh needed in: %" APR_TIME_T_FMT " seconds", apr_time_sec(last_refresh + interval - apr_time_now()));
 
@@ -1008,23 +1002,31 @@ static int oidc_handle_existing_session(request_rec *r, oidc_cfg *cfg,
 	if ((r->user != NULL) && (dir_cfg->authn_header != NULL))
 		oidc_util_set_header(r, dir_cfg->authn_header, r->user);
 
-	/* set the claims in the app headers + request state */
-	if (oidc_set_app_claims(r, cfg, session, OIDC_CLAIMS_SESSION_KEY) == FALSE)
+	const char *s_claims = NULL;
+	const char *s_id_token = NULL;
+
+	/* get the string-encoded claims JSON object from the session */
+	oidc_session_get(r, session, OIDC_CLAIMS_SESSION_KEY, &s_claims);
+	/* set the claims JSON string in the request state so it is available for authz purposes later on */
+	oidc_request_state_set(r, OIDC_CLAIMS_SESSION_KEY, s_claims);
+
+	/* set the claims in the app headers  */
+	if (oidc_set_app_claims(r, cfg, session, s_claims) == FALSE)
 		return HTTP_INTERNAL_SERVER_ERROR;
 
+	/* get the string-encoded id_token JSON object from the session */
+	oidc_session_get(r, session, OIDC_IDTOKEN_CLAIMS_SESSION_KEY, &s_id_token);
+	/* set the claims JSON string in the request state so it is available for authz purposes later on */
+	oidc_request_state_set(r, OIDC_IDTOKEN_CLAIMS_SESSION_KEY, s_id_token);
+
 	if ((cfg->pass_idtoken_as & OIDC_PASS_IDTOKEN_AS_CLAIMS)) {
-		/* set the id_token in the app headers + request state */
-		if (oidc_set_app_claims(r, cfg, session,
-				OIDC_IDTOKEN_CLAIMS_SESSION_KEY) == FALSE)
+		/* set the id_token in the app headers */
+		if (oidc_set_app_claims(r, cfg, session, s_id_token) == FALSE)
 			return HTTP_INTERNAL_SERVER_ERROR;
 	}
 
 	if ((cfg->pass_idtoken_as & OIDC_PASS_IDTOKEN_AS_PAYLOAD)) {
-		const char *s_id_token = NULL;
-		/* get the string-encoded JSON object from the session */
-		oidc_session_get(r, session, OIDC_IDTOKEN_CLAIMS_SESSION_KEY,
-				&s_id_token);
-		/* pass it to the app in a header or environment variable */
+		/* pass the id_token JSON object to the app in a header or environment variable */
 		oidc_util_set_app_info(r, "id_token_payload", s_id_token,
 				OIDC_DEFAULT_HEADER_PREFIX, dir_cfg->pass_info_in_headers,
 				dir_cfg->pass_info_in_env_vars);
@@ -1034,7 +1036,7 @@ static int oidc_handle_existing_session(request_rec *r, oidc_cfg *cfg,
 		const char *s_id_token = NULL;
 		/* get the compact serialized JWT from the session */
 		oidc_session_get(r, session, OIDC_IDTOKEN_SESSION_KEY, &s_id_token);
-		/* pass it to the app in a header or environment variable */
+		/* pass the compact serialized JWT to the app in a header or environment variable */
 		oidc_util_set_app_info(r, "id_token", s_id_token,
 				OIDC_DEFAULT_HEADER_PREFIX, dir_cfg->pass_info_in_headers,
 				dir_cfg->pass_info_in_env_vars);
@@ -1190,7 +1192,8 @@ static int oidc_authorization_response_error(request_rec *r, oidc_cfg *c,
  * set the unique user identifier that will be propagated in the Apache r->user and REMOTE_USER variables
  */
 static apr_byte_t oidc_get_remote_user(request_rec *r, oidc_cfg *c,
-		oidc_provider_t *provider, apr_jwt_t *jwt, char **user) {
+		oidc_provider_t *provider, apr_jwt_t *jwt, char **user,
+		const char *s_claims) {
 
 	char *issuer = provider->issuer;
 	char *claim_name = apr_pstrdup(r->pool, c->remote_user_claim.claim_name);
@@ -1204,12 +1207,21 @@ static apr_byte_t oidc_get_remote_user(request_rec *r, oidc_cfg *c,
 						apr_pstrdup(r->pool, issuer + strlen("https://"));
 	}
 
-	/* extract the username claim (default: "sub") from the id_token payload */
+	/* extract the username claim (default: "sub") from the id_token payload or user claims */
 	char *username = NULL;
-	if (apr_jwt_get_string(r->pool, jwt->payload.value.json, claim_name, TRUE,
-			&username, NULL) == FALSE) {
+	json_error_t json_error;
+	json_t *claims = json_loads(s_claims, 0, &json_error);
+	if (claims == NULL) {
+		username = apr_pstrdup(r->pool, json_string_value(json_object_get(jwt->payload.value.json, claim_name)));
+	} else {
+		oidc_util_json_merge(jwt->payload.value.json, claims);
+		username = apr_pstrdup(r->pool, json_string_value(json_object_get(claims, claim_name)));
+		json_decref(claims);
+	}
+
+	if (username == NULL) {
 		oidc_error(r,
-				"OIDCRemoteUserClaim is set to \"%s\", but the id_token JSON payload did not contain a \"%s\" string",
+				"OIDCRemoteUserClaim is set to \"%s\", but the id_token JSON payload and user claims did not contain a \"%s\" string",
 				c->remote_user_claim.claim_name, claim_name);
 		*user = NULL;
 		return FALSE;
@@ -1218,7 +1230,7 @@ static apr_byte_t oidc_get_remote_user(request_rec *r, oidc_cfg *c,
 	/* set the unique username in the session (will propagate to r->user/REMOTE_USER) */
 	*user = post_fix_with_issuer ?
 			apr_psprintf(r->pool, "%s@%s", username, issuer) :
-			apr_pstrdup(r->pool, username);
+			username;
 
 	if (c->remote_user_claim.reg_exp != NULL) {
 
@@ -1489,7 +1501,7 @@ static int oidc_handle_authorization_response(request_rec *r, oidc_cfg *c,
 			json_string_value(json_object_get(proto_state, "original_method")));
 
 	/* set the user */
-	if (oidc_get_remote_user(r, c, provider, jwt, &r->user) == TRUE) {
+	if (oidc_get_remote_user(r, c, provider, jwt, &r->user, claims) == TRUE) {
 
 		/* session management: if the user in the new response is not equal to the old one, error out */
 		if ((json_object_get(proto_state, "prompt") != NULL)
@@ -2502,6 +2514,11 @@ int oidc_handle_redirect_uri_request(request_rec *r, oidc_cfg *c,
  */
 static int oidc_check_userid_openidc(request_rec *r, oidc_cfg *c) {
 
+	if (c->redirect_uri == NULL) {
+		oidc_error(r, "configuration error: the authentication type is set to \"openid-connect\" but OIDCRedirectURI has not been set");
+		return HTTP_INTERNAL_SERVER_ERROR;
+	}
+
 	/* check if this is a sub-request or an initial request */
 	if (ap_is_initial_req(r)) {
 
@@ -2557,9 +2574,12 @@ static int oidc_check_userid_openidc(request_rec *r, oidc_cfg *c) {
 
 	/* find out which action we need to take when encountering an unauthenticated request */
 	switch (dir_cfg->unauth_action) {
+		case RETURN410:
+			return HTTP_GONE;
 		case RETURN401:
 			return HTTP_UNAUTHORIZED;
 		case PASS:
+			r->user = "";
 			return OK;
 		case AUTHENTICATE:
 			/* if this is a Javascript path we won't redirect the user and create a state cookie */
@@ -2633,6 +2653,15 @@ static void oidc_authz_get_claims_and_idtoken(request_rec *r, json_t **claims,
  */
 authz_status oidc_authz_checker(request_rec *r, const char *require_args, const void *parsed_require_args) {
 
+	/* check for anonymous access and PASS mode */
+	if (r->user != NULL && strlen(r->user) == 0) {
+		r->user = NULL;
+		/* get a handle to the directory config */
+		oidc_dir_cfg *dir_cfg = ap_get_module_config(r->per_dir_config,
+				&auth_openidc_module);
+		if (dir_cfg->unauth_action == PASS) return AUTHZ_GRANTED;
+	}
+
 	/* get the set of claims from the request state (they've been set in the authentication part earlier */
 	json_t *claims = NULL, *id_token = NULL;
 	oidc_authz_get_claims_and_idtoken(r, &claims, &id_token);
@@ -2648,6 +2677,11 @@ authz_status oidc_authz_checker(request_rec *r, const char *require_args, const
 	if (claims) json_decref(claims);
 	if (id_token) json_decref(id_token);
 
+	if ((rc == HTTP_UNAUTHORIZED) && ap_auth_type(r)
+			&& (apr_strnatcasecmp((const char *) ap_auth_type(r), "oauth20")
+					== 0))
+		oidc_oauth_return_www_authenticate(r, "insufficient_scope", "Different scope(s) or other claims required");
+
 	return rc;
 }
 #else
@@ -2657,6 +2691,15 @@ authz_status oidc_authz_checker(request_rec *r, const char *require_args, const
  */
 int oidc_auth_checker(request_rec *r) {
 
+	/* check for anonymous access and PASS mode */
+	if (r->user != NULL && strlen(r->user) == 0) {
+		r->user = NULL;
+		/* get a handle to the directory config */
+		oidc_dir_cfg *dir_cfg = ap_get_module_config(r->per_dir_config,
+				&auth_openidc_module);
+		if (dir_cfg->unauth_action == PASS) return OK;
+	}
+
 	/* get the set of claims from the request state (they've been set in the authentication part earlier */
 	json_t *claims = NULL, *id_token = NULL;
 	oidc_authz_get_claims_and_idtoken(r, &claims, &id_token);
@@ -2687,6 +2730,11 @@ int oidc_auth_checker(request_rec *r) {
 	if (id_token)
 		json_decref(id_token);
 
+	if ((rc == HTTP_UNAUTHORIZED) && ap_auth_type(r)
+			&& (apr_strnatcasecmp((const char *) ap_auth_type(r), "oauth20")
+					== 0))
+		oidc_oauth_return_www_authenticate(r, "insufficient_scope", "Different scope(s) or other claims required");
+
 	return rc;
 }
 #endif
