@@ -18,7 +18,7 @@
  */
 
 /***************************************************************************
- * Copyright (C) 2013-2015 Ping Identity Corporation
+ * Copyright (C) 2013-2016 Ping Identity Corporation
  * All rights reserved.
  *
  * For further information please contact:
@@ -109,6 +109,10 @@ APLOG_USE_MODULE(auth_openidc);
 #define OIDC_REFRESHTOKEN_SESSION_KEY "refresh_token"
 /* key for storing maximum session duration in the session context */
 #define OIDC_SESSION_EXPIRES_SESSION_KEY "session_expires"
+/* key for storing the cookie domain in the session context */
+#define OIDC_COOKIE_DOMAIN_SESSION_KEY "cookie_domain"
+/* key for storing last user info refresh timestamp in the session context */
+#define OIDC_USERINFO_LAST_REFRESH_SESSION_KEY "userinfo_last_refresh"
 /* key for storing request state */
 #define OIDC_REQUEST_STATE_SESSION_KEY "request_state"
 /* key for storing the original URL */
@@ -144,6 +148,9 @@ APLOG_USE_MODULE(auth_openidc);
 /* nonce bytes length */
 #define OIDC_PROTO_NONCE_LENGTH 32
 
+/* code verifier length */
+#define OIDC_PROTO_CODE_VERIFIER_LENGTH 32
+
 /* pass id_token as individual claims in headers (default) */
 #define OIDC_PASS_IDTOKEN_AS_CLAIMS     1
 /* pass id_token payload as JSON object in header*/
@@ -151,6 +158,16 @@ APLOG_USE_MODULE(auth_openidc);
 /* pass id_token in compact serialized format in header*/
 #define OIDC_PASS_IDTOKEN_AS_SERIALIZED 4
 
+#define OIDC_OAUTH_ACCEPT_TOKEN_IN_DEFAULT 0
+/* accept bearer token in header (default) */
+#define OIDC_OAUTH_ACCEPT_TOKEN_IN_HEADER  1
+/* accept bearer token as a post parameter */
+#define OIDC_OAUTH_ACCEPT_TOKEN_IN_POST    2
+/* accept bearer token as a query parameter */
+#define OIDC_OAUTH_ACCEPT_TOKEN_IN_QUERY   4
+/* accept bearer token as a cookie parameter (PingAccess) */
+#define OIDC_OAUTH_ACCEPT_TOKEN_IN_COOKIE  8
+
 /* prefix of the cookie that binds the state in the authorization request/response to the browser */
 #define OIDCStateCookiePrefix  "mod_auth_openidc_state_"
 
@@ -186,7 +203,7 @@ APLOG_USE_MODULE(auth_openidc);
 #define OIDC_CACHE_SECTION_PROVIDER "provider"
 
 typedef enum {
-	AUTHENTICATE, PASS, RETURN401
+	AUTHENTICATE, PASS, RETURN401, RETURN410
 } unauthenticated_action;
 
 typedef struct oidc_jwks_uri_t {
@@ -223,6 +240,8 @@ typedef struct oidc_provider_t {
 	int idtoken_iat_slack;
 	char *auth_request_params;
 	int session_max_duration;
+	char *pkce_method;
+	int userinfo_refresh_interval;
 
 	char *client_jwks_uri;
 	char *id_token_signed_response_alg;
@@ -260,6 +279,9 @@ typedef struct oidc_cfg {
 	/* indicates whether this is a derived config, merged from a base one */
 	unsigned int merged;
 
+	/* HTML to display error messages+description */
+	char *error_template;
+
 	/* the redirect URI as configured with the OpenID Connect OP's that we talk to */
 	char *redirect_uri;
 	/* (optional) default URL for 3rd-party initiated SSO */
@@ -282,6 +304,8 @@ typedef struct oidc_cfg {
 	char *metadata_dir;
 	/* type of session management/storage */
 	int session_type;
+	/* session cookie or persistent cookie */
+	int persistent_session_cookie;
 
 	/* pointer to cache functions */
 	oidc_cache_t *cache;
@@ -335,6 +359,9 @@ typedef struct oidc_dir_cfg {
 	apr_array_header_t *pass_cookies;
 	apr_byte_t pass_info_in_headers;
 	apr_byte_t pass_info_in_env_vars;
+	apr_byte_t oauth_accept_token_in;
+	apr_hash_t *oauth_accept_token_options;
+	int oauth_token_introspect_interval;
 } oidc_dir_cfg;
 
 int oidc_check_user_id(request_rec *r);
@@ -352,10 +379,9 @@ int oidc_oauth_check_userid(request_rec *r, oidc_cfg *c);
 
 // oidc_proto.c
 
-int oidc_proto_authorization_request(request_rec *r, struct oidc_provider_t *provider, const char *login_hint, const char *redirect_uri, const char *state, json_t *proto_state, const char *id_token_hint, const char *auth_request_params);
+int oidc_proto_authorization_request(request_rec *r, struct oidc_provider_t *provider, const char *login_hint, const char *redirect_uri, const char *state, json_t *proto_state, const char *id_token_hint, const char *code_challenge, const char *auth_request_params);
 apr_byte_t oidc_proto_is_post_authorization_response(request_rec *r, oidc_cfg *cfg);
 apr_byte_t oidc_proto_is_redirect_authorization_response(request_rec *r, oidc_cfg *cfg);
-apr_byte_t oidc_proto_resolve_code(request_rec *r, oidc_cfg *cfg, oidc_provider_t *provider, const char *code, char **id_token, char **access_token, char **token_type, int *expires_in, char **refresh_token);
 apr_byte_t oidc_proto_refresh_request(request_rec *r, oidc_cfg *cfg, oidc_provider_t *provider, const char *rtoken, char **id_token, char **access_token, char **token_type, int *expires_in, char **refresh_token);
 apr_byte_t oidc_proto_resolve_userinfo(request_rec *r, oidc_cfg *cfg, oidc_provider_t *provider, const char *access_token, const char **response);
 apr_byte_t oidc_proto_account_based_discovery(request_rec *r, oidc_cfg *cfg, const char *acct, char **issuer);
@@ -367,6 +393,8 @@ apr_byte_t oidc_proto_validate_authorization_response(request_rec *r, const char
 apr_byte_t oidc_proto_jwt_verify(request_rec *r, oidc_cfg *cfg, apr_jwt_t *jwt, const oidc_jwks_uri_t *jwks_uri, apr_hash_t *symmetric_keys);
 apr_byte_t oidc_proto_validate_jwt(request_rec *r, apr_jwt_t *jwt, const char *iss, apr_byte_t exp_is_mandatory, apr_byte_t iat_is_mandatory, int iat_slack);
 apr_byte_t oidc_proto_generate_nonce(request_rec *r, char **nonce, int len);
+apr_byte_t oidc_proto_generate_code_verifier(request_rec *r, char **code_verifier, int len);
+apr_byte_t oidc_proto_generate_code_challenge(request_rec *r, const char *code_verifier, char **code_challenge, const char *challenge_method);
 
 apr_byte_t oidc_proto_authorization_response_code_idtoken_token(request_rec *r, oidc_cfg *c, json_t *proto_state, oidc_provider_t *provider, apr_table_t *params, const char *response_mode, apr_jwt_t **jwt);
 apr_byte_t oidc_proto_authorization_response_code_idtoken(request_rec *r, oidc_cfg *c, json_t *proto_state, oidc_provider_t *provider, apr_table_t *params, const char *response_mode, apr_jwt_t **jwt);
@@ -385,6 +413,7 @@ int oidc_authz_worker(request_rec *r, const json_t *const claims, const require_
 #if MODULE_MAGIC_NUMBER_MAJOR >= 20100714
 authz_status oidc_authz_worker24(request_rec *r, const json_t * const claims, const char *require_line);
 #endif
+int oidc_oauth_return_www_authenticate(request_rec *r, const char *error, const char *error_description);
 
 // oidc_config.c
 void *oidc_create_server_config(apr_pool_t *pool, server_rec *svr);
@@ -399,7 +428,8 @@ int oidc_base64url_encode(request_rec *r, char **dst, const char *src, int src_l
 int oidc_base64url_decode(request_rec *r, char **dst, const char *src);
 int oidc_encrypt_base64url_encode_string(request_rec *r, char **dst, const char *src);
 int oidc_base64url_decode_decrypt_string(request_rec *r, char **dst, const char *src);
-char *oidc_get_current_url(const request_rec *r, const oidc_cfg *c);
+const char *oidc_get_current_url_host(request_rec *r);
+char *oidc_get_current_url(request_rec *r);
 char *oidc_url_encode(const request_rec *r, const char *str, const char *charsToEncode);
 char *oidc_normalize_header_name(const request_rec *r, const char *str);
 
@@ -418,15 +448,16 @@ char *oidc_util_escape_string(const request_rec *r, const char *str);
 char *oidc_util_unescape_string(const request_rec *r, const char *str);
 apr_byte_t oidc_util_read_form_encoded_params(request_rec *r, apr_table_t *table, const char *data);
 apr_byte_t oidc_util_read_post_params(request_rec *r, apr_table_t *table);
-apr_byte_t oidc_util_file_read(request_rec *r, const char *path, char **result);
+apr_byte_t oidc_util_file_read(request_rec *r, const char *path, apr_pool_t *pool, char **result);
 apr_byte_t oidc_util_issuer_match(const char *a, const char *b);
-int oidc_util_html_send_error(request_rec *r, const char *error, const char *description, int status_code);
+int oidc_util_html_send_error(request_rec *r, const char *html_template, const char *error, const char *description, int status_code);
 apr_byte_t oidc_util_json_array_has_value(request_rec *r, json_t *haystack, const char *needle);
+void oidc_util_set_header(request_rec *r, const char *s_name, const char *s_value);
 void oidc_util_set_app_info(request_rec *r, const char *s_key, const char *s_value, const char *claim_prefix, apr_byte_t as_header, apr_byte_t as_env_var);
 void oidc_util_set_app_infos(request_rec *r, const json_t *j_attrs, const char *claim_prefix, const char *claim_delimiter, apr_byte_t as_header, apr_byte_t as_env_var);
 apr_hash_t *oidc_util_spaced_string_to_hashtable(apr_pool_t *pool, const char *str);
 apr_byte_t oidc_util_spaced_string_equals(apr_pool_t *pool, const char *a, const char *b);
-apr_byte_t oidc_util_spaced_string_contains(apr_pool_t *pool, const char *response_type, const char *match);
+apr_byte_t oidc_util_spaced_string_contains(apr_pool_t *pool, const char *str, const char *match);
 apr_byte_t oidc_json_object_get_string(apr_pool_t *pool, json_t *json, const char *name, char **value, const char *default_value);
 apr_byte_t oidc_json_object_get_int(apr_pool_t *pool, json_t *json, const char *name, int *value, const int default_value);
 char *oidc_util_html_escape(apr_pool_t *pool, const char *input);
@@ -435,6 +466,8 @@ apr_hash_t * oidc_util_merge_symmetric_key(apr_pool_t *pool, apr_hash_t *private
 apr_hash_t * oidc_util_merge_key_sets(apr_pool_t *pool, apr_hash_t *k1, apr_hash_t *k2);
 apr_byte_t oidc_util_regexp_first_match(apr_pool_t *pool, const char *input, const char *regexp, char **output, char **error_str);
 apr_byte_t oidc_util_json_merge(json_t *src, json_t *dst);
+int oidc_util_cookie_domain_valid(const char *hostname, char *cookie_domain);
+apr_byte_t oidc_util_hash_string_and_base64url_encode(request_rec *r, const char *openssl_hash_algo, const char *input, char **output);
 
 // oidc_crypto.c
 unsigned char *oidc_crypto_aes_encrypt(request_rec *r, oidc_cfg *cfg, unsigned char *plaintext, int *len);
