@@ -18,7 +18,7 @@
  */
 
 /***************************************************************************
- * Copyright (C) 2013-2015 Ping Identity Corporation
+ * Copyright (C) 2013-2016 Ping Identity Corporation
  * All rights reserved.
  *
  * For further information please contact:
@@ -111,32 +111,84 @@ static apr_byte_t oidc_oauth_validate_access_token(request_rec *r, oidc_cfg *c,
 static apr_byte_t oidc_oauth_get_bearer_token(request_rec *r,
 		const char **access_token) {
 
-	/* get the authorization header */
-	const char *auth_line;
-	auth_line = apr_table_get(r->headers_in, "Authorization");
-	if (!auth_line) {
-		oidc_debug(r, "no authorization header found");
+	/* get a handle to the directory config */
+	oidc_dir_cfg *dir_cfg = ap_get_module_config(r->per_dir_config,
+			&auth_openidc_module);
+
+	*access_token = NULL;
+
+	if ((dir_cfg->oauth_accept_token_in & OIDC_OAUTH_ACCEPT_TOKEN_IN_HEADER)
+			|| (dir_cfg->oauth_accept_token_in
+					== OIDC_OAUTH_ACCEPT_TOKEN_IN_DEFAULT)) {
+
+		/* get the authorization header */
+		const char *auth_line;
+		auth_line = apr_table_get(r->headers_in, "Authorization");
+		if (auth_line) {
+			oidc_debug(r, "authorization header found");
+
+			/* look for the Bearer keyword */
+			if (apr_strnatcasecmp(ap_getword(r->pool, &auth_line, ' '),
+					"Bearer") == 0) {
+
+				/* skip any spaces after the Bearer keyword */
+				while (apr_isspace(*auth_line)) {
+					auth_line++;
+				}
+
+				/* copy the result in to the access_token */
+				*access_token = apr_pstrdup(r->pool, auth_line);
+
+			} else {
+				oidc_warn(r,
+						"client used unsupported authentication scheme: %s",
+						r->uri);
+			}
+		}
+	}
+
+	if ((*access_token == NULL) && (r->method_number == M_POST)
+			&& (dir_cfg->oauth_accept_token_in & OIDC_OAUTH_ACCEPT_TOKEN_IN_POST)) {
+		apr_table_t *params = apr_table_make(r->pool, 8);
+		if (oidc_util_read_post_params(r, params) == TRUE) {
+			*access_token = apr_table_get(params, "access_token");
+		}
+	}
+
+	if ((*access_token == NULL)
+			&& (dir_cfg->oauth_accept_token_in
+					& OIDC_OAUTH_ACCEPT_TOKEN_IN_QUERY)) {
+		apr_table_t *params = apr_table_make(r->pool, 8);
+		oidc_util_read_form_encoded_params(r, params, r->args);
+		*access_token = apr_table_get(params, "access_token");
+	}
+
+	if ((*access_token == NULL)
+			&& (dir_cfg->oauth_accept_token_in
+					& OIDC_OAUTH_ACCEPT_TOKEN_IN_COOKIE)) {
+
+		const char *cookie_name = apr_hash_get(
+				dir_cfg->oauth_accept_token_options, "cookie-name",
+				APR_HASH_KEY_STRING);
+		const char *auth_line = oidc_util_get_cookie(r, cookie_name);
+		if (auth_line != NULL) {
+
+			/* copy the result in to the access_token */
+			*access_token = apr_pstrdup(r->pool, auth_line);
+
+		} else {
+			oidc_warn(r, "no cookie found with name: %s", cookie_name);
+		}
+	}
+
+	if (*access_token == NULL) {
+		oidc_debug(r,
+				"no bearer token found in the allowed methods (authorization header, post, query parameter or cookie)");
 		return FALSE;
 	}
 
-	/* look for the Bearer keyword */
-	if (apr_strnatcasecmp(ap_getword(r->pool, &auth_line, ' '), "Bearer")) {
-		oidc_error(r, "client used unsupported authentication scheme: %s",
-				r->uri);
-		return FALSE;
-	}
-
-	/* skip any spaces after the Bearer keyword */
-	while (apr_isspace(*auth_line)) {
-		auth_line++;
-	}
-
-	/* copy the result in to the access_token */
-	*access_token = apr_pstrdup(r->pool, auth_line);
-
 	/* log some stuff */
 	oidc_debug(r, "bearer token: %s", *access_token);
-
 	return TRUE;
 }
 
@@ -219,6 +271,83 @@ static apr_byte_t oidc_oauth_parse_and_cache_token_expiry(request_rec *r,
 	return TRUE;
 }
 
+static apr_byte_t oidc_oauth_cache_access_token(request_rec *r, oidc_cfg *c,
+		apr_time_t cache_until, const char *access_token, json_t *json) {
+
+	oidc_debug(r, "caching introspection result");
+
+	json_t *cache_entry = json_object();
+	json_object_set(cache_entry, "response", json);
+	json_object_set_new(cache_entry, "timestamp",
+			json_integer(apr_time_sec(apr_time_now())));
+	char *cache_value = json_dumps(cache_entry, 0);
+
+	/* set it in the cache so subsequent request don't need to validate the access_token and get the claims anymore */
+	c->cache->set(r, OIDC_CACHE_SECTION_ACCESS_TOKEN, access_token, cache_value,
+			cache_until);
+
+	json_decref(cache_entry);
+	free(cache_value);
+
+	return TRUE;
+}
+
+static apr_byte_t oidc_oauth_get_cached_access_token(request_rec *r,
+		oidc_cfg *c, const char *access_token, json_t **json) {
+	json_t *cache_entry = NULL;
+	const char *s_cache_entry = NULL;
+	json_error_t json_error;
+
+	oidc_dir_cfg *dir_cfg = ap_get_module_config(r->per_dir_config,
+			&auth_openidc_module);
+
+	/* see if we've got the claims for this access_token cached already */
+	c->cache->get(r, OIDC_CACHE_SECTION_ACCESS_TOKEN, access_token,
+			&s_cache_entry);
+
+	if (s_cache_entry == NULL)
+		return FALSE;
+
+	/* json decode the cache entry */
+	cache_entry = json_loads(s_cache_entry, 0, &json_error);
+	if (cache_entry == NULL) {
+		oidc_error(r, "cached JSON was corrupted: %s", json_error.text);
+		*json = NULL;
+		return FALSE;
+	}
+
+	/* compare the timestamp against the freshness requirement */
+	json_t *v = json_object_get(cache_entry, "timestamp");
+	apr_time_t now = apr_time_sec(apr_time_now());
+	if ((dir_cfg->oauth_token_introspect_interval > 0)
+			&& (now
+					> json_integer_value(v)
+					+ dir_cfg->oauth_token_introspect_interval)) {
+
+		/* printout info about the event */
+		char buf[APR_RFC822_DATE_LEN + 1];
+		apr_rfc822_date(buf, apr_time_from_sec(json_integer_value(v)));
+		oidc_debug(r,
+				"token that was validated/cached at: [%s], does not meet token freshness requirement: %d)",
+				buf, dir_cfg->oauth_token_introspect_interval);
+
+		/* invalidate the cache entry */
+		*json = NULL;
+		json_decref(cache_entry);
+		return FALSE;
+	}
+
+	oidc_debug(r,
+			"returning cached introspection result that meets freshness requirements: %s",
+			s_cache_entry);
+
+	/* we've got a cached introspection result that is still valid for this path's requirements */
+	*json = json_deep_copy(json_object_get(cache_entry, "response"));
+
+	json_decref(cache_entry);
+	return TRUE;
+}
+
 /*
  * resolve and validate an access_token against the configured Authorization Server
  */
@@ -226,22 +355,24 @@ static apr_byte_t oidc_oauth_resolve_access_token(request_rec *r, oidc_cfg *c,
 		const char *access_token, json_t **token, char **response) {
 
 	json_t *result = NULL;
-	const char *json = NULL;
 
 	/* see if we've got the claims for this access_token cached already */
-	c->cache->get(r, OIDC_CACHE_SECTION_ACCESS_TOKEN, access_token, &json);
+	oidc_oauth_get_cached_access_token(r, c, access_token, &result);
 
-	if (json == NULL) {
+	if (result == NULL) {
+
+		const char *s_json = NULL;
 
 		/* not cached, go out and validate the access_token against the Authorization server and get the JSON claims back */
-		if (oidc_oauth_validate_access_token(r, c, access_token, &json) == FALSE) {
+		if (oidc_oauth_validate_access_token(r, c, access_token,
+				&s_json) == FALSE) {
 			oidc_error(r,
 					"could not get a validation response from the Authorization server");
 			return FALSE;
 		}
 
 		/* decode and see if it is not an error response somehow */
-		if (oidc_util_decode_json_and_check_error(r, json, &result) == FALSE)
+		if (oidc_util_decode_json_and_check_error(r, s_json, &result) == FALSE)
 			return FALSE;
 
 		json_t *active = json_object_get(result, "active");
@@ -277,8 +408,8 @@ static apr_byte_t oidc_oauth_resolve_access_token(request_rec *r, oidc_cfg *c,
 			}
 
 			/* set it in the cache so subsequent request don't need to validate the access_token and get the claims anymore */
-			c->cache->set(r, OIDC_CACHE_SECTION_ACCESS_TOKEN, access_token,
-					json, cache_until);
+			oidc_oauth_cache_access_token(r, c, cache_until, access_token,
+					result);
 
 		} else {
 
@@ -294,20 +425,11 @@ static apr_byte_t oidc_oauth_resolve_access_token(request_rec *r, oidc_cfg *c,
 			}
 
 			/* set it in the cache so subsequent request don't need to validate the access_token and get the claims anymore */
-			c->cache->set(r, OIDC_CACHE_SECTION_ACCESS_TOKEN, access_token,
-					json, cache_until);
+			oidc_oauth_cache_access_token(r, c, cache_until, access_token,
+					result);
 
 		}
 
-	} else {
-
-		/* we got the claims for this access_token in our cache, decode it in to a JSON structure */
-		json_error_t json_error;
-		result = json_loads(json, 0, &json_error);
-		if (result == NULL) {
-			oidc_error(r, "cached JSON was corrupted: %s", json_error.text);
-			return FALSE;
-		}
 	}
 
 	/* return the access_token JSON object */
@@ -325,9 +447,6 @@ static apr_byte_t oidc_oauth_resolve_access_token(request_rec *r, oidc_cfg *c,
 
 		/* return only the pimped access_token results */
 		*token = json_deep_copy(tkn);
-		char *s_token = json_dumps(*token, 0);
-		*response = apr_pstrdup(r->pool, s_token);
-		free(s_token);
 
 		json_decref(result);
 
@@ -337,10 +456,13 @@ static apr_byte_t oidc_oauth_resolve_access_token(request_rec *r, oidc_cfg *c,
 
 		/* assume spec compliant introspection */
 		*token = result;
-		*response = apr_pstrdup(r->pool, json);
 
 	}
 
+	char *s_token = json_dumps(*token, 0);
+	*response = apr_pstrdup(r->pool, s_token);
+	free(s_token);
+
 	return TRUE;
 }
 
@@ -404,6 +526,8 @@ static apr_byte_t oidc_oauth_validate_jwt_access_token(request_rec *r,
 		return FALSE;
 	}
 
+	oidc_debug(r, "successfully parsed JWT with header: %s", jwt->header.value.str);
+
 	/* validate the access token JWT, validating optional exp + iat */
 	if (oidc_proto_validate_jwt(r, jwt, NULL, FALSE, FALSE,
 			c->provider.idtoken_iat_slack) == FALSE) {
@@ -439,6 +563,21 @@ static apr_byte_t oidc_oauth_validate_jwt_access_token(request_rec *r,
 	return TRUE;
 }
 
+/*
+ * set the WWW-Authenticate response header according to https://tools.ietf.org/html/rfc6750#section-3
+ */
+int oidc_oauth_return_www_authenticate(request_rec *r, const char *error, const char *error_description) {
+	char *hdr = apr_psprintf(r->pool, "Bearer");
+	if (ap_auth_name(r) != NULL)
+		hdr = apr_psprintf(r->pool, "%s realm=\"%s\"", hdr, ap_auth_name(r));
+	if (error != NULL)
+		hdr = apr_psprintf(r->pool, "%s%s error=\"%s\"", hdr, (ap_auth_name(r) ? "," : ""), error);
+	if (error_description != NULL)
+		hdr = apr_psprintf(r->pool, "%s, error_description=\"%s\"", hdr, error_description);
+	apr_table_setn(r->err_headers_out, "WWW-Authenticate", hdr);
+	return HTTP_UNAUTHORIZED;
+}
+
 /*
  * main routine: handle OAuth 2.0 authentication/authorization
  */
@@ -479,7 +618,7 @@ int oidc_oauth_check_userid(request_rec *r, oidc_cfg *c) {
 	/* get the bearer access token from the Authorization header */
 	const char *access_token = NULL;
 	if (oidc_oauth_get_bearer_token(r, &access_token) == FALSE)
-		return HTTP_UNAUTHORIZED;
+		return oidc_oauth_return_www_authenticate(r, "invalid_request", "No bearer token found in the request");
 
 	/* validate the obtained access token against the OAuth AS validation endpoint */
 	json_t *token = NULL;
@@ -490,18 +629,18 @@ int oidc_oauth_check_userid(request_rec *r, oidc_cfg *c) {
 		/* we'll validate the token remotely */
 		if (oidc_oauth_resolve_access_token(r, c, access_token, &token,
 				&s_token) == FALSE)
-			return HTTP_UNAUTHORIZED;
+			return oidc_oauth_return_www_authenticate(r, "invalid_token", "Reference token could not be introspected");
 	} else {
 		/* no introspection endpoint is set, assume the token is a JWT and validate it locally */
 		if (oidc_oauth_validate_jwt_access_token(r, c, access_token, &token,
 				&s_token) == FALSE)
-			return HTTP_UNAUTHORIZED;
+			return oidc_oauth_return_www_authenticate(r, "invalid_token", "JWT token could not be validated");
 	}
 
 	/* check that we've got something back */
 	if (token == NULL) {
 		oidc_error(r, "could not resolve claims (token == NULL)");
-		return HTTP_UNAUTHORIZED;
+		return oidc_oauth_return_www_authenticate(r, "invalid_token", "No claims could be parsed from the token");
 	}
 
 	/* store the parsed token (cq. the claims from the response) in the request state so it can be accessed by the authz routines */
@@ -511,7 +650,7 @@ int oidc_oauth_check_userid(request_rec *r, oidc_cfg *c) {
 	if (oidc_oauth_set_remote_user(r, c, token) == FALSE) {
 		oidc_error(r,
 				"remote user could not be set, aborting with HTTP_UNAUTHORIZED");
-		return HTTP_UNAUTHORIZED;
+		return oidc_oauth_return_www_authenticate(r, "invalid_token", "Could not set remote user");
 	}
 
 	/* get a handle to the director config */
