@@ -18,7 +18,7 @@
  */
 
 /***************************************************************************
- * Copyright (C) 2013-2015 Ping Identity Corporation
+ * Copyright (C) 2013-2016 Ping Identity Corporation
  * All rights reserved.
  *
  * For further information please contact:
@@ -103,12 +103,14 @@ int oidc_proto_authorization_request_post_preserve(request_rec *r,
 int oidc_proto_authorization_request(request_rec *r,
 		struct oidc_provider_t *provider, const char *login_hint,
 		const char *redirect_uri, const char *state, json_t *proto_state,
-		const char *id_token_hint, const char *auth_request_params) {
+		const char *id_token_hint, const char *code_challenge,
+		const char *auth_request_params) {
 
 	/* log some stuff */
 	char *s_value = json_dumps(proto_state, JSON_ENCODE_ANY);
-	oidc_debug(r, "enter, issuer=%s, redirect_uri=%s, state=%s, proto_state=%s",
-			provider->issuer, redirect_uri, state, s_value);
+	oidc_debug(r,
+			"enter, issuer=%s, redirect_uri=%s, state=%s, proto_state=%s, code_challenge=%s",
+			provider->issuer, redirect_uri, state, s_value, code_challenge);
 	free(s_value);
 
 	/* assemble the full URL as the authorization request to the OP where we want to redirect to */
@@ -121,8 +123,21 @@ int oidc_proto_authorization_request(request_rec *r,
 			oidc_util_escape_string(r,
 					json_string_value(
 							json_object_get(proto_state, "response_type"))));
-	authorization_request = apr_psprintf(r->pool, "%s&scope=%s",
-			authorization_request, oidc_util_escape_string(r, provider->scope));
+
+	if (provider->scope != NULL) {
+
+		if (!oidc_util_spaced_string_contains(r->pool, provider->scope,
+				"openid")) {
+			oidc_warn(r,
+					"the configuration for the \"scope\" parameter does not include the \"openid\" scope, your provider may not return an \"id_token\": %s",
+					provider->scope);
+		}
+
+		authorization_request = apr_psprintf(r->pool, "%s&scope=%s",
+				authorization_request,
+				oidc_util_escape_string(r, provider->scope));
+	}
+
 	authorization_request = apr_psprintf(r->pool, "%s&client_id=%s",
 			authorization_request,
 			oidc_util_escape_string(r, provider->client_id));
@@ -139,6 +154,14 @@ int oidc_proto_authorization_request(request_rec *r,
 						json_string_value(
 								json_object_get(proto_state, "nonce"))));
 
+	/* add PKCE code challenge if set */
+	if (code_challenge != NULL)
+		authorization_request = apr_psprintf(r->pool,
+				"%s&code_challenge=%s&code_challenge_method=%s",
+				authorization_request,
+				oidc_util_escape_string(r, code_challenge),
+				provider->pkce_method);
+
 	/* add the response_mode if explicitly set */
 	if (json_object_get(proto_state, "response_mode") != NULL)
 		authorization_request = apr_psprintf(r->pool, "%s&response_mode=%s",
@@ -226,13 +249,14 @@ apr_byte_t oidc_proto_is_redirect_authorization_response(request_rec *r,
 /*
  * generate a random value (nonce) to correlate request/response through browser state
  */
-apr_byte_t oidc_proto_generate_nonce(request_rec *r, char **nonce, int len) {
-	unsigned char *nonce_bytes = apr_pcalloc(r->pool, len);
-	if (apr_generate_random_bytes(nonce_bytes, len) != APR_SUCCESS) {
+static apr_byte_t oidc_proto_generate_random_string(request_rec *r,
+		char **output, int len) {
+	unsigned char *bytes = apr_pcalloc(r->pool, len);
+	if (apr_generate_random_bytes(bytes, len) != APR_SUCCESS) {
 		oidc_error(r, "apr_generate_random_bytes returned an error");
 		return FALSE;
 	}
-	if (oidc_base64url_encode(r, nonce, (const char *) nonce_bytes, len, TRUE)
+	if (oidc_base64url_encode(r, output, (const char *) bytes, len, TRUE)
 			<= 0) {
 		oidc_error(r, "oidc_base64url_encode returned an error");
 		return FALSE;
@@ -240,6 +264,52 @@ apr_byte_t oidc_proto_generate_nonce(request_rec *r, char **nonce, int len) {
 	return TRUE;
 }
 
+/*
+ * generate a random value (nonce) to correlate request/response through browser state
+ */
+apr_byte_t oidc_proto_generate_nonce(request_rec *r, char **nonce, int len) {
+	return oidc_proto_generate_random_string(r, nonce, len);
+}
+
+/*
+ * generate a random value (code_verifier) to correlate authorization request and token exchange request through browser state
+ */
+apr_byte_t oidc_proto_generate_code_verifier(request_rec *r,
+		char **code_verifier, int len) {
+	//*code_verifier = apr_pstrdup(r->pool, "dBjftJeZ4CVP-mB92K27uhbUJU1p1r_wW1gFWFOEjXk"); return TRUE;
+	return oidc_proto_generate_random_string(r, code_verifier, len);
+}
+
+/*
+ * generate the PKCE code challenge for a given code verifier and method
+ */
+apr_byte_t oidc_proto_generate_code_challenge(request_rec *r,
+		const char *code_verifier, char **code_challenge,
+		const char *challenge_method) {
+
+	oidc_debug(r, "enter: method=%s", challenge_method);
+
+	if (code_verifier != NULL) {
+
+		if (apr_strnatcmp(challenge_method, "plain") == 0) {
+
+			*code_challenge = apr_pstrdup(r->pool, code_verifier);
+
+		} else if (apr_strnatcmp(challenge_method, "S256") == 0) {
+
+			if (oidc_util_hash_string_and_base64url_encode(r, "sha256",
+					code_verifier, code_challenge) == FALSE) {
+				oidc_error(r,
+						"oidc_util_hash_string_and_base64url_encode returned an error for the code verifier");
+				return FALSE;
+			}
+		}
+
+	}
+
+	return TRUE;
+}
+
 /*
  * if a nonce was passed in the authorization request (and stored in the browser state),
  * check that it matches the nonce value in the id_token payload
@@ -565,7 +635,8 @@ static apr_byte_t oidc_proto_get_key_from_jwks(request_rec *r, apr_jwt_t *jwt,
 			else if (jwk->kid)
 				apr_hash_set(result, jwk->kid, APR_HASH_KEY_STRING, jwk);
 			else
-				apr_hash_set(result, apr_psprintf(r->pool, "%d", i), APR_HASH_KEY_STRING, jwk);
+				apr_hash_set(result, apr_psprintf(r->pool, "%d", i),
+						APR_HASH_KEY_STRING, jwk);
 			continue;
 		}
 
@@ -794,16 +865,106 @@ static apr_byte_t oidc_proto_token_endpoint_request(request_rec *r,
 
 	const char *response = NULL;
 
-	/* see if we need to do basic auth or auth-through-post-params (both applied through the HTTP POST method though) */
 	const char *basic_auth = NULL;
-	if ((provider->token_endpoint_auth == NULL)
-			|| (apr_strnatcmp(provider->token_endpoint_auth,
-					"client_secret_basic") == 0)) {
-		basic_auth = apr_psprintf(r->pool, "%s:%s", provider->client_id,
-				provider->client_secret);
+	if (provider->client_secret != NULL) {
+		/* see if we need to do basic auth or auth-through-post-params (both applied through the HTTP POST method though) */
+		if ((provider->token_endpoint_auth == NULL)
+				|| (apr_strnatcmp(provider->token_endpoint_auth,
+						"client_secret_basic") == 0)) {
+			basic_auth = apr_psprintf(r->pool, "%s:%s", provider->client_id,
+					provider->client_secret);
+		} else if ((provider->token_endpoint_auth != NULL)
+				&& ((apr_strnatcmp(provider->token_endpoint_auth,
+						"client_secret_jwt") == 0) || (apr_strnatcmp(provider->token_endpoint_auth,
+								"private_key_jwt") == 0))) {
+
+			apr_jwt_t jwt;
+			jwt.header.kid = NULL;
+			jwt.header.alg = NULL;
+			jwt.header.enc = NULL;
+			jwt.header.value.json = json_object();
+			json_object_set_new(jwt.header.value.json, "typ",
+					json_string("JWT"));
+
+			char *jti = NULL;
+			oidc_proto_generate_random_string(r, &jti, 16);
+			jwt.payload.value.json = json_object();
+			json_object_set_new(jwt.payload.value.json, "iss",
+					json_string(provider->client_id));
+			json_object_set_new(jwt.payload.value.json, "sub",
+					json_string(provider->client_id));
+			json_object_set_new(jwt.payload.value.json, "aud",
+					json_string(provider->token_endpoint_url));
+			json_object_set_new(jwt.payload.value.json, "jti",
+					json_string(jti));
+			json_object_set_new(jwt.payload.value.json, "exp",
+					json_integer(apr_time_sec(apr_time_now()) + 60));
+			json_object_set_new(jwt.payload.value.json, "iat",
+					json_integer(apr_time_sec(apr_time_now())));
+
+			apr_jwk_t *jwk = NULL;
+			apr_jwt_error_t err;
+
+			if (apr_strnatcmp(provider->token_endpoint_auth,
+					"client_secret_jwt") == 0) {
+
+				if (apr_jwk_parse_symmetric_key(r->pool, NULL,
+						(const unsigned char *) provider->client_secret,
+						strlen(provider->client_secret), &jwk, &err) == FALSE) {
+					oidc_error(r,
+							"parsing of client secret into JWK failed: %s",
+							apr_jwt_e2s(r->pool, err));
+					apr_jwt_destroy(&jwt);
+					return FALSE;
+				}
+
+				json_object_set_new(jwt.header.value.json, "alg",
+						json_string("HS256"));
+
+			} else {
+
+				if (cfg->private_keys == NULL) {
+					oidc_error(r,
+							"no private keys have been configured to use for private_key_jwt client authentication (OIDCPrivateKeyFiles)");
+					apr_jwt_destroy(&jwt);
+					return FALSE;
+				}
+
+				char *kid = NULL;
+				apr_ssize_t klen = 0;
+				apr_hash_index_t *hi = apr_hash_first(r->pool,
+						cfg->private_keys);
+				apr_hash_this(hi, (const void **) &kid, &klen, (void **) &jwk);
+
+				json_object_set_new(jwt.header.value.json, "alg",
+						json_string("RS256"));
+				json_object_set_new(jwt.header.value.json, "kid",
+						json_string(kid));
+
+			}
+
+			if (apr_jwt_sign(r->pool, &jwt, jwk, &err) == FALSE) {
+				oidc_error(r, "signing JWT failed: %s",
+						apr_jwt_e2s(r->pool, err));
+				apr_jwt_destroy(&jwt);
+				return FALSE;
+			}
+
+			apr_table_addn(params, "client_assertion_type",
+					"urn:ietf:params:oauth:client-assertion-type:jwt-bearer");
+			apr_table_addn(params, "client_assertion",
+					apr_jwt_serialize(r->pool, &jwt));
+
+			apr_jwt_destroy(&jwt);
+
+		} else {
+			apr_table_addn(params, "client_id", provider->client_id);
+			apr_table_addn(params, "client_secret", provider->client_secret);
+		}
 	} else {
+		oidc_debug(r,
+				"no client secret is configured; calling the token endpoint without client authentication; only public clients are supported");
 		apr_table_addn(params, "client_id", provider->client_id);
-		apr_table_addn(params, "client_secret", provider->client_secret);
 	}
 
 	/* add any configured extra static parameters to the token endpoint */
@@ -859,10 +1020,10 @@ static apr_byte_t oidc_proto_token_endpoint_request(request_rec *r,
 /*
  * resolves the code received from the OP in to an id_token, access_token and refresh_token
  */
-apr_byte_t oidc_proto_resolve_code(request_rec *r, oidc_cfg *cfg,
-		oidc_provider_t *provider, const char *code, char **id_token,
-		char **access_token, char **token_type, int *expires_in,
-		char **refresh_token) {
+static apr_byte_t oidc_proto_resolve_code(request_rec *r, oidc_cfg *cfg,
+		oidc_provider_t *provider, const char *code, const char *code_verifier,
+		char **id_token, char **access_token, char **token_type,
+		int *expires_in, char **refresh_token, const char *state) {
 
 	oidc_debug(r, "enter");
 
@@ -872,6 +1033,12 @@ apr_byte_t oidc_proto_resolve_code(request_rec *r, oidc_cfg *cfg,
 	apr_table_addn(params, "code", code);
 	apr_table_addn(params, "redirect_uri", cfg->redirect_uri);
 
+	if (code_verifier)
+		apr_table_addn(params, "code_verifier", code_verifier);
+
+	if (state)
+		apr_table_addn(params, "state", state);
+
 	return oidc_proto_token_endpoint_request(r, cfg, provider, params, id_token,
 			access_token, token_type, expires_in, refresh_token);
 }
@@ -1327,17 +1494,52 @@ static apr_byte_t oidc_proto_validate_response_mode(request_rec *r,
 	return TRUE;
 }
 
+/*
+ * validate the client_id/iss provided by the OP against the client_id/iss registered with the provider that the request was sent to
+ */
+static apr_byte_t oidc_proto_validate_issuer_client_id(request_rec *r,
+		const char *configured_issuer, const char *response_issuer,
+		const char *configured_client_id, const char *response_client_id) {
+
+	if (response_issuer != NULL) {
+		if (apr_strnatcmp(configured_issuer, response_issuer) != 0) {
+			oidc_error(r,
+					"configured issuer (%s) does not match the issuer provided in the response by the OP (%s)",
+					configured_issuer, response_issuer);
+			return FALSE;
+		}
+	}
+
+	if (response_client_id != NULL) {
+		if (apr_strnatcmp(configured_client_id, response_client_id) != 0) {
+			oidc_error(r,
+					"configured client_id (%s) does not match the client_id provided in the response by the OP (%s)",
+					configured_client_id, response_client_id);
+			return FALSE;
+		}
+	}
+
+	oidc_debug(r, "iss and/or client_id matched OK: %s, %s, %s, %s", response_issuer, configured_issuer, response_client_id, configured_client_id);
+
+	return TRUE;
+}
+
 /*
  * helper function to validate both the response type and the response mode in a single function call
  */
-static apr_byte_t oidc_proto_validate_response_type_and_response_mode(
-		request_rec *r, const char *requested_response_type,
-		apr_table_t *params, json_t *proto_state, const char *response_mode,
-		const char *default_response_mode) {
+static apr_byte_t oidc_proto_validate_response_type_mode_issuer(request_rec *r,
+		const char *requested_response_type, apr_table_t *params,
+		json_t *proto_state, const char *response_mode,
+		const char *default_response_mode, const char *issuer, const char *c_client_id) {
 
 	const char *code = apr_table_get(params, "code");
 	const char *id_token = apr_table_get(params, "id_token");
 	const char *access_token = apr_table_get(params, "access_token");
+	const char *iss = apr_table_get(params, "iss");
+	const char *client_id = apr_table_get(params, "client_id");
+
+	if (oidc_proto_validate_issuer_client_id(r, issuer, iss, c_client_id, client_id) == FALSE)
+		return FALSE;
 
 	if (oidc_proto_validate_response_type(r, requested_response_type, code,
 			id_token, access_token) == FALSE)
@@ -1385,7 +1587,7 @@ static apr_byte_t oidc_proto_parse_idtoken_and_validate_code(request_rec *r,
  */
 static apr_byte_t oidc_proto_resolve_code_and_validate_response(request_rec *r,
 		oidc_cfg *c, oidc_provider_t *provider, const char *response_type,
-		apr_table_t *params) {
+		apr_table_t *params, json_t *proto_state) {
 
 	char *id_token = NULL;
 	char *access_token = NULL;
@@ -1393,9 +1595,21 @@ static apr_byte_t oidc_proto_resolve_code_and_validate_response(request_rec *r,
 	int expires_in = -1;
 	char *refresh_token = NULL;
 
+	const char *code_verifier =
+			json_object_get(proto_state, "code_verifier") ?
+					json_string_value(
+							json_object_get(proto_state, "code_verifier")) :
+					NULL;
+
+	const char *state =
+			json_object_get(proto_state, "state") ?
+					json_string_value(
+							json_object_get(proto_state, "state")) :
+					NULL;
+
 	if (oidc_proto_resolve_code(r, c, provider, apr_table_get(params, "code"),
-			&id_token, &access_token, &token_type, &expires_in,
-			&refresh_token) == FALSE) {
+			code_verifier, &id_token, &access_token, &token_type, &expires_in,
+			&refresh_token, state) == FALSE) {
 		oidc_error(r, "failed to resolve the code");
 		return FALSE;
 	}
@@ -1440,8 +1654,9 @@ apr_byte_t oidc_proto_authorization_response_code_idtoken(request_rec *r,
 
 	static const char *response_type = "code id_token";
 
-	if (oidc_proto_validate_response_type_and_response_mode(r, response_type,
-			params, proto_state, response_mode, "fragment") == FALSE)
+	if (oidc_proto_validate_response_type_mode_issuer(r, response_type, params,
+			proto_state, response_mode, "fragment",
+			provider->issuer, provider->client_id) == FALSE)
 		return FALSE;
 
 	if (oidc_proto_parse_idtoken_and_validate_code(r, c, proto_state, provider,
@@ -1455,7 +1670,7 @@ apr_byte_t oidc_proto_authorization_response_code_idtoken(request_rec *r,
 	apr_table_unset(params, "refresh_token");
 
 	if (oidc_proto_resolve_code_and_validate_response(r, c, provider,
-			response_type, params) == FALSE)
+			response_type, params, proto_state) == FALSE)
 		return FALSE;
 
 	return TRUE;
@@ -1472,8 +1687,9 @@ apr_byte_t oidc_proto_handle_authorization_response_code_token(request_rec *r,
 
 	static const char *response_type = "code token";
 
-	if (oidc_proto_validate_response_type_and_response_mode(r, response_type,
-			params, proto_state, response_mode, "fragment") == FALSE)
+	if (oidc_proto_validate_response_type_mode_issuer(r, response_type, params,
+			proto_state, response_mode, "fragment",
+			provider->issuer, provider->client_id) == FALSE)
 		return FALSE;
 
 	/* clear parameters that should only be set from the token endpoint */
@@ -1481,7 +1697,7 @@ apr_byte_t oidc_proto_handle_authorization_response_code_token(request_rec *r,
 	apr_table_unset(params, "refresh_token");
 
 	if (oidc_proto_resolve_code_and_validate_response(r, c, provider,
-			response_type, params) == FALSE)
+			response_type, params, proto_state) == FALSE)
 		return FALSE;
 
 	if (oidc_proto_parse_idtoken_and_validate_code(r, c, proto_state, provider,
@@ -1502,8 +1718,8 @@ apr_byte_t oidc_proto_handle_authorization_response_code(request_rec *r,
 
 	static const char *response_type = "code";
 
-	if (oidc_proto_validate_response_type_and_response_mode(r, response_type,
-			params, proto_state, response_mode, "query") == FALSE)
+	if (oidc_proto_validate_response_type_mode_issuer(r, response_type, params,
+			proto_state, response_mode, "query", provider->issuer, provider->client_id) == FALSE)
 		return FALSE;
 
 	/* clear parameters that should only be set from the token endpoint */
@@ -1514,7 +1730,7 @@ apr_byte_t oidc_proto_handle_authorization_response_code(request_rec *r,
 	apr_table_unset(params, "refresh_token");
 
 	if (oidc_proto_resolve_code_and_validate_response(r, c, provider,
-			response_type, params) == FALSE)
+			response_type, params, proto_state) == FALSE)
 		return FALSE;
 
 	/*
@@ -1544,8 +1760,9 @@ static apr_byte_t oidc_proto_handle_implicit_flow(request_rec *r, oidc_cfg *c,
 		oidc_provider_t *provider, apr_table_t *params,
 		const char *response_mode, apr_jwt_t **jwt) {
 
-	if (oidc_proto_validate_response_type_and_response_mode(r, response_type,
-			params, proto_state, response_mode, "fragment") == FALSE)
+	if (oidc_proto_validate_response_type_mode_issuer(r, response_type, params,
+			proto_state, response_mode, "fragment",
+			provider->issuer, provider->client_id) == FALSE)
 		return FALSE;
 
 	if (oidc_proto_parse_idtoken_and_validate_code(r, c, proto_state, provider,
@@ -1578,7 +1795,7 @@ apr_byte_t oidc_proto_authorization_response_code_idtoken_token(request_rec *r,
 	apr_table_unset(params, "refresh_token");
 
 	if (oidc_proto_resolve_code_and_validate_response(r, c, provider,
-			response_type, params) == FALSE)
+			response_type, params, proto_state) == FALSE)
 		return FALSE;
 
 	return TRUE;
