@@ -18,7 +18,7 @@
  */
 
 /***************************************************************************
- * Copyright (C) 2013-2015 Ping Identity Corporation
+ * Copyright (C) 2013-2016 Ping Identity Corporation
  * All rights reserved.
  *
  * For further information please contact:
@@ -315,45 +315,67 @@ static const char *oidc_get_current_url_scheme(const request_rec *r) {
  * get the URL port that is currently being accessed
  */
 static const char *oidc_get_current_url_port(const request_rec *r,
-		const oidc_cfg *c, const char *scheme_str) {
+		const char *scheme_str) {
 	/* first see if there's a proxy/load-balancer in front of us */
 	const char *port_str = apr_table_get(r->headers_in, "X-Forwarded-Port");
 	if (port_str == NULL) {
-		/* if not we'll take the port from the Host header (as set by the client or ProxyPreserveHost) */
-		const char *host_hdr = apr_table_get(r->headers_in, "Host");
-		port_str = strchr(host_hdr, ':');
-		if (port_str == NULL) {
-			/* if no port was set in the Host header we'll determine it locally */
-			const apr_port_t port = r->connection->local_addr->port;
-			apr_byte_t print_port = TRUE;
-			if ((apr_strnatcmp(scheme_str, "https") == 0) && port == 443)
-				print_port = FALSE;
-			else if ((apr_strnatcmp(scheme_str, "http") == 0) && port == 80)
-				print_port = FALSE;
-			if (print_port)
-				port_str = apr_psprintf(r->pool, "%u", port);
+		/* see if we can get the port from the "X-Forwarded-Host" header */
+		const char *host_hdr = apr_table_get(r->headers_in, "X-Forwarded-Host");
+		if (host_hdr) {
+			port_str = strchr(host_hdr, ':');
+			if (port_str)
+				port_str++;
 		} else {
-			port_str++;
+			host_hdr = apr_table_get(r->headers_in, "Host");
+			if (host_hdr)
+				port_str = strchr(host_hdr, ':');
+			if (port_str == NULL) {
+				/* if no port was set in the Host header we'll determine it locally */
+				const apr_port_t port = r->connection->local_addr->port;
+				apr_byte_t print_port = TRUE;
+				if ((apr_strnatcmp(scheme_str, "https") == 0) && port == 443)
+					print_port = FALSE;
+				else if ((apr_strnatcmp(scheme_str, "http") == 0) && port == 80)
+					print_port = FALSE;
+				if (print_port)
+					port_str = apr_psprintf(r->pool, "%u", port);
+			} else {
+				port_str++;
+			}
 		}
 	}
 	return port_str;
 }
 
+/*
+ * get the hostname part of the URL that is currently being accessed
+ */
+const char *oidc_get_current_url_host(request_rec *r) {
+	const char *host_str = apr_table_get(r->headers_in, "X-Forwarded-Host");
+	if (host_str == NULL)
+		host_str = apr_table_get(r->headers_in, "Host");
+	if (host_str) {
+		host_str = apr_pstrdup(r->pool, host_str);
+		char *p = strchr(host_str, ':');
+		if (p != NULL)
+			*p = '\0';
+	} else {
+		/* no Host header, HTTP 1.0 */
+		host_str = ap_get_server_name(r);
+	}
+	return host_str;
+}
+
 /*
  * get the URL that is currently being accessed
  */
-char *oidc_get_current_url(const request_rec *r, const oidc_cfg *c) {
+char *oidc_get_current_url(request_rec *r) {
 
 	const char *scheme_str = oidc_get_current_url_scheme(r);
-
-	const char *port_str = oidc_get_current_url_port(r, c, scheme_str);
+	const char *host_str = oidc_get_current_url_host(r);
+	const char *port_str = oidc_get_current_url_port(r, scheme_str);
 	port_str = port_str ? apr_psprintf(r->pool, ":%s", port_str) : "";
 
-	const char *host_str = apr_table_get(r->headers_in, "Host");
-	char *p = strchr(host_str, ':');
-	if (p != NULL)
-		*p = '\0';
-
 	char *url = apr_pstrcat(r->pool, scheme_str, "://", host_str, port_str,
 			r->uri, (r->args != NULL && *r->args != '\0' ? "?" : ""), r->args,
 			NULL);
@@ -633,7 +655,7 @@ static char *oidc_util_get_path(request_rec *r) {
 	size_t i;
 	char *p;
 	p = r->parsed_uri.path;
-	if (p[0] == '\0')
+	if ((p == NULL) || (p[0] == '\0'))
 		return apr_pstrdup(r->pool, "/");
 	for (i = strlen(p) - 1; i > 0; i--)
 		if (p[i] == '/')
@@ -697,6 +719,13 @@ void oidc_util_set_cookie(request_rec *r, const char *cookieName,
 					";Secure" : ""),
 			c->cookie_http_only != FALSE ? ";HttpOnly" : "");
 
+	/* sanity check on overall cookie value size */
+	if (strlen(headerString) > 4093) {
+		oidc_warn(r,
+				"the length of the cookie value (%lu) is greater than 4093(!) bytes, this may not work with all browsers/server combinations: consider switching to a server side caching!",
+				(unsigned long)strlen(headerString));
+	}
+
 	/* use r->err_headers_out so we always print our headers (even on 302 redirect) - headers_out only prints on 2xx responses */
 	apr_table_add(r->err_headers_out, "Set-Cookie", headerString);
 
@@ -733,7 +762,7 @@ char *oidc_util_get_cookie(request_rec *r, const char *cookieName) {
 				cookie++;
 
 			/* see if we've found the cookie that we're looking for */
-			if (strncmp(cookie, cookieName, strlen(cookieName)) == 0) {
+			if ((strncmp(cookie, cookieName, strlen(cookieName)) == 0)  && (cookie[strlen(cookieName)] == '=')) {
 
 				/* skip to the meat of the parameter (the value after the '=') */
 				cookie += (strlen(cookieName) + 1);
@@ -946,24 +975,49 @@ int oidc_util_html_send(request_rec *r, const char *title,
 	return oidc_util_http_send(r, html, strlen(html), "text/html", status_code);
 }
 
+static char *html_error_template_contents = NULL;
+
 /*
  * send a user-facing error to the browser
  */
-int oidc_util_html_send_error(request_rec *r, const char *error,
-		const char *description, int status_code) {
+int oidc_util_html_send_error(request_rec *r, const char *html_template,
+		const char *error, const char *description, int status_code) {
 
-	char *html_body = "";
+	char *html = "";
+
+	if (html_template != NULL) {
+
+		if (html_error_template_contents == NULL) {
+			int rc = oidc_util_file_read(r, html_template,
+					r->server->process->pool, &html_error_template_contents);
+			if (rc == FALSE) {
+				oidc_error(r, "could not read HTML error template: %s",
+						html_template);
+				html_error_template_contents = NULL;
+			}
+		}
+
+		if (html_error_template_contents) {
+			html = apr_psprintf(r->pool, html_error_template_contents,
+					oidc_util_html_escape(r->pool, error ? error : ""),
+					oidc_util_html_escape(r->pool,
+							description ? description : ""));
+
+			return oidc_util_http_send(r, html, strlen(html), "text/html",
+					status_code);
+		}
+	}
 
 	if (error != NULL) {
-		html_body = apr_psprintf(r->pool, "%s<p>Error: <pre>%s</pre></p>",
-				html_body, oidc_util_html_escape(r->pool, error));
+		html = apr_psprintf(r->pool, "%s<p>Error: <pre>%s</pre></p>", html,
+				oidc_util_html_escape(r->pool, error));
 	}
 	if (description != NULL) {
-		html_body = apr_psprintf(r->pool, "%s<p>Description: <pre>%s</pre></p>",
-				html_body, oidc_util_html_escape(r->pool, description));
+		html = apr_psprintf(r->pool, "%s<p>Description: <pre>%s</pre></p>",
+				html, oidc_util_html_escape(r->pool, description));
 	}
 
-	return oidc_util_html_send(r, "Error", NULL, NULL, html_body, status_code);
+	return oidc_util_html_send(r, "Error", NULL, NULL, html, status_code);
 }
 
 /*
@@ -1034,15 +1088,14 @@ apr_byte_t oidc_util_read_post_params(request_rec *r, apr_table_t *table) {
 /*
  * read a file from a path on disk
  */
-apr_byte_t oidc_util_file_read(request_rec *r, const char *path, char **result) {
+apr_byte_t oidc_util_file_read(request_rec *r, const char *path, apr_pool_t *pool, char **result) {
 	apr_file_t *fd = NULL;
 	apr_status_t rc = APR_SUCCESS;
 	char s_err[128];
 	apr_finfo_t finfo;
 
 	/* open the file if it exists */
-	if ((rc = apr_file_open(&fd, path, APR_FOPEN_READ | APR_FOPEN_BUFFERED,
-	APR_OS_DEFAULT, r->pool)) != APR_SUCCESS) {
+	if ((rc = apr_file_open(&fd, path, APR_FOPEN_READ | APR_FOPEN_BUFFERED, APR_OS_DEFAULT, r->pool)) != APR_SUCCESS) {
 		oidc_warn(r, "no file found at: \"%s\"", path);
 		return FALSE;
 	}
@@ -1062,7 +1115,7 @@ apr_byte_t oidc_util_file_read(request_rec *r, const char *path, char **result)
 	}
 
 	/* now that we have the size of the file, allocate a buffer that can contain its contents */
-	*result = apr_palloc(r->pool, finfo.size + 1);
+	*result = apr_palloc(pool, finfo.size + 1);
 
 	/* read the file in to the buffer */
 	apr_size_t bytes_read = 0;
@@ -1109,7 +1162,7 @@ error_close:
 apr_byte_t oidc_util_issuer_match(const char *a, const char *b) {
 
 	/* check the "issuer" value against the one configure for the provider we got this id_token from */
-	if (strcmp(a, b) != 0) {
+	if (apr_strnatcmp(a, b) != 0) {
 
 		/* no strict match, but we are going to accept if the difference is only a trailing slash */
 		int n1 = strlen(a);
@@ -1140,7 +1193,7 @@ apr_byte_t oidc_util_json_array_has_value(request_rec *r, json_t *haystack,
 					elem->type);
 			continue;
 		}
-		if (strcmp(json_string_value(elem), needle) == 0) {
+		if (apr_strnatcmp(json_string_value(elem), needle) == 0) {
 			break;
 		}
 	}
@@ -1153,7 +1206,31 @@ apr_byte_t oidc_util_json_array_has_value(request_rec *r, json_t *haystack,
 }
 
 /*
- * set an HTTP header to pass information to the application
+ * set a HTTP header
+ */
+void oidc_util_set_header(request_rec *r, const char *s_name, const char *s_value) {
+	/*
+	 * sanitize the header value by replacing line feeds with spaces
+	 * just like the Apache header input algorithms do for incoming headers
+	 *
+	 * this makes it impossible to have line feeds in values but that is
+	 * compliant with RFC 7230 (and impossible for regular headers due to Apache's
+	 * parsing of headers anyway) and fixes a security vulnerability on
+	 * overwriting/setting outgoing headers when used in proxy mode
+	 */
+	char *p = NULL;
+	while ((p = strchr(s_value, '\n')))
+		*p = ' ';
+
+	/* do some logging about this event */
+	oidc_debug(r, "setting header \"%s: %s\"", s_name, s_value);
+
+	/* now set the actual header name/value */
+	apr_table_set(r->headers_in, s_name, s_value);
+}
+
+/*
+ * set a HTTP header and/or environment variable to pass information to the application
  */
 void oidc_util_set_app_info(request_rec *r, const char *s_key,
 		const char *s_value, const char *claim_prefix, apr_byte_t as_header,
@@ -1165,26 +1242,8 @@ void oidc_util_set_app_info(request_rec *r, const char *s_key,
 	const char *s_name = apr_psprintf(r->pool, "%s%s", claim_prefix,
 			oidc_normalize_header_name(r, s_key));
 
-	if (as_header) {
-		/*
-		 * sanitize the header value by replacing line feeds with spaces
-		 * just like the Apache header input algorithms do for incoming headers
-		 *
-		 * this makes it impossible to have line feeds in values but that is
-		 * compliant with RFC 7230 (and impossible for regular headers due to Apache's
-		 * parsing of headers anyway) and fixes a security vulnerability on
-		 * overwriting/setting outgoing headers when used in proxy mode
-		 */
-		char *p = NULL;
-		while ((p = strchr(s_value, '\n')))
-			*p = ' ';
-
-		/* do some logging about this event */
-		oidc_debug(r, "setting header \"%s: %s\"", s_name, s_value);
-
-		/* now set the actual header name/value */
-		apr_table_set(r->headers_in, s_name, s_value);
-	}
+	if (as_header)
+		oidc_util_set_header(r, s_name, s_value);
 
 	if (as_env_var) {
 
@@ -1383,8 +1442,8 @@ apr_byte_t oidc_util_spaced_string_equals(apr_pool_t *pool, const char *a,
  * see if a particular value is part of a space separated value
  */
 apr_byte_t oidc_util_spaced_string_contains(apr_pool_t *pool,
-		const char *response_type, const char *match) {
-	apr_hash_t *ht = oidc_util_spaced_string_to_hashtable(pool, response_type);
+		const char *str, const char *match) {
+	apr_hash_t *ht = oidc_util_spaced_string_to_hashtable(pool, str);
 	return (apr_hash_get(ht, match, APR_HASH_KEY_STRING) != NULL);
 }
 
@@ -1482,7 +1541,8 @@ apr_hash_t * oidc_util_merge_symmetric_key(apr_pool_t *pool, apr_hash_t *keys,
 					strlen(client_secret), &key, &key_len, &err);
 		}
 
-		if (apr_jwk_parse_symmetric_key(pool, key, key_len, &jwk, &err) == TRUE) {
+		if (apr_jwk_parse_symmetric_key(pool, NULL, key, key_len, &jwk,
+				&err) == TRUE) {
 			apr_hash_set(result, jwk->kid, APR_HASH_KEY_STRING, jwk);
 		}
 	}
@@ -1490,6 +1550,31 @@ apr_hash_t * oidc_util_merge_symmetric_key(apr_pool_t *pool, apr_hash_t *keys,
 	return result;
 }
 
+/*
+ * openssl hash and base64 encode
+ */
+apr_byte_t oidc_util_hash_string_and_base64url_encode(request_rec *r,
+		const char *openssl_hash_algo, const char *input, char **output) {
+	apr_jwt_error_t err;
+	unsigned char *hashed = NULL;
+	unsigned int hashed_len = 0;
+	if (apr_jws_hash_bytes(r->pool, openssl_hash_algo,
+			(const unsigned char *) input, strlen(input), &hashed, &hashed_len,
+			&err) == FALSE) {
+		oidc_error(r,
+				"apr_jws_hash_bytes returned an error: %s", err.text);
+		return FALSE;
+	}
+
+	if (oidc_base64url_encode(r, output, (const char *) hashed, hashed_len,
+			TRUE) <= 0) {
+		oidc_error(r,
+				"oidc_base64url_encode returned an error: %s", err.text);
+		return FALSE;
+	}
+	return TRUE;
+}
+
 /*
  * merge two key sets
  */
@@ -1575,3 +1660,19 @@ apr_byte_t oidc_util_regexp_first_match(apr_pool_t *pool, const char *input,
 
 	return TRUE;
 }
+
+int oidc_util_cookie_domain_valid(const char *hostname, char *cookie_domain) {
+	char *p = NULL;
+	char *check_cookie = cookie_domain;
+	// Skip past the first char of a cookie_domain that starts
+	// with a ".", ASCII 46
+	if (check_cookie[0] == 46) {
+		p = strstr(hostname, ++check_cookie);
+	} else {
+		p = strstr(hostname, check_cookie);
+	}
+	if ((p == NULL) || (apr_strnatcmp(check_cookie, p) != 0)) {
+		return FALSE;
+	}
+	return TRUE;
+}
